# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
MicrodiffAperture. Move the aperture in the beam to a specified value or
out of the beam.
The factor, which serves to calculate the flux, can be a single value or a
tuple of values per aperture size.

Example xml file:
<object class="MicrodiffAperture">
  <username>aperture</username>
  <exporter_address>wid30bmd2s:9001</exporter_address>
  <value_channel_name>CurrentApertureDiameterIndex</value_channel_name>
  <state_channel_name>State</state_channel_name>
  <-- either only factor -->
  <factor>(0.15, 0.3, 0.63, 0.9, 0.96)</factor>
  <!-- or complete, corresponding to label: (index, size[um], factor) -->
  <values>{"A10": (0, 10, 0.15), "A20": (1, 20, 0.3), "A30": (2, 30, 0.63), "A50": (3, 50, 0.9), "A75": (4, 75, 0.96)}</values>
  <object role="inout" href="/udiff_apertureinout"/>
</object>
"""

from ast import literal_eval
from enum import Enum

from mxcubecore.HardwareObjects.abstract.AbstractNState import BaseValueEnum
from mxcubecore.HardwareObjects.ExporterNState import ExporterNState

__copyright__ = """ Copyright © 2022 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class MicrodiffAperture(ExporterNState):
    """MicrodiffAperture class"""

    unit = "um"

    def __init__(self, name):
        super().__init__(name)
        self.inout_obj = None

    def init(self):
        """Initialize the aperture"""
        super().init()

        # check if we have values other that UNKNOWN (no values in config)
        if len(self.VALUES) == 1:
            self._initialise_values()

        # now get the IN/OUT object
        self.inout_obj = self.get_object_by_role("inout")
        if self.inout_obj:
            self._initialise_inout()

    def _set_value(self, value):
        """Set device to value
        Args:
            value (str, int, float or enum): Value to be set.
        """
        if value.name in ("IN", "OUT"):
            _eval = self.inout_obj.value_to_enum(value.value)
            self.inout_obj.set_value(_eval, timeout=60)
        else:
            super()._set_value(value)
            # put the aperture in
            if self.inout_obj:
                self.inout_obj.set_value(self.inout_obj.VALUES.IN, timeout=60)

    def _initialise_inout(self):
        """Add IN and OUT to the values Enum"""
        values_dict = {item.name: item.value for item in self.inout_obj.VALUES}
        values_dict.update({item.name: item.value for item in self.VALUES})
        self.VALUES = Enum("ValueEnum", values_dict)

    def _initialise_values(self):
        """Initialise the ValueEnum from the hardware
        Raises:
            RuntimeError: No aperture diameters defined.
                          Factor and aperture diameter not the same number.
                          Invalid factor values.
        """
        predefined_postions = self._exporter.read_property("ApertureDiameters")
        if not predefined_postions:
            raise RuntimeError("No aperture diameters defined")

        values = {}
        try:
            # get the factors
            factor = literal_eval(self.get_property("factor"))
            if len(predefined_postions) == len(factor):
                for _pos, _fac in zip(predefined_postions, factor):
                    values["A{0}".format(_pos)] = (
                        predefined_postions.index(_pos),
                        _pos,
                        _fac,
                    )
            else:
                raise RuntimeError("Factor and aperture diameter not the same number")
        except (ValueError, TypeError):
            raise RuntimeError("Invalid factor values")

        self.VALUES = Enum(
            "ValueEnum",
            dict(values, **{item.name: item.value for item in BaseValueEnum}),
        )

    def get_factor(self, label):
        """Get the factor associated to a label.
        Args:
            (enum, str): label enum or name
        Returns:
            (float) or (tuple): Factor value
        """
        if isinstance(label, str):
            try:
                return self.VALUES[label].value[2]
            except (KeyError, ValueError, IndexError):
                return 1.0
        try:
            return label.value[2]
        except (ValueError, IndexError):
            return 1.0

    def get_size(self, label):
        """Get the aperture size associated to a label.
        Args:
            (enum, str): label enum or name
        Returns:
            (float): Factor value
        Raises:
            RuntimeError: Unknown aperture size.
        """
        if isinstance(label, str):
            try:
                return float(self.VALUES[label].value[1])
            except (KeyError, ValueError, IndexError):
                raise RuntimeError("Unknown aperture size")
        try:
            return float(label.value[1])
        except (ValueError, IndexError):
            if self.inout_obj:
                return None
            raise RuntimeError("Unknown aperture size")

    def get_diameter_size_list(self):
        """Get the list of values to be visible. Hide IN, OUT and UNKNOWN.
        Returns:
            (list): List of available aperture values (string).
        """
        values = []
        for value in self.VALUES:
            _nam = value.name

            if _nam not in ["IN", "OUT", "UNKNOWN"]:
                values.append(_nam)

        return values



import logging
import time

from mxcubecore.HardwareObjects.LNLS.LNLSMotor import LNLSMotor


class LNLSDetDistMotor(LNLSMotor):
    PMAC_SENDCMD = "pmac_sendcmd"

    def __init__(self, name):
        LNLSMotor.__init__(self, name)

    def _move(self, value):
        """Override super class method."""
        if not self.validate_value(value):
            raise ValueError(
                "Invalid value %s; limits are %s" % (value, self.get_limits())
            )
        self.update_specific_state(self.SPECIFIC_STATES.MOVING)

        for i in range(2):
            logging.getLogger("user_level_log").info("Setting detector distance...")

        # Enable air
        logging.getLogger("HWR").info("%s: Enabling air" % self.motor_name)
        command = "#5,7,8j/"
        self.set_channel_value(self.PMAC_SENDCMD, command)
        time.sleep(2)

        # Wait for stability
        logging.getLogger("HWR").info("%s: Wait for air stability" % self.motor_name)
        while self.get_channel_value(self.MOTOR_DMOV) == 0:
            time.sleep(0.2)
            current_value = self.get_value()
            self.update_value(current_value)

        # Move det dist motor
        logging.getLogger("HWR").info("%s: Set motor to %s" % (self.motor_name, value))
        self.set_channel_value(self.ACTUATOR_VAL, value)

        # Wait for movement
        logging.getLogger("HWR").info(
            "%s: Wait for movement to finish" % self.motor_name
        )
        while (
            self.get_channel_value(self.MOTOR_DMOV) == 0
            or abs(current_value - value) > 0.0005
        ):
            time.sleep(0.2)
            current_value = self.get_value()
            self.update_value(current_value)

        # Disable air
        time.sleep(1)
        logging.getLogger("HWR").info("%s: Disabling air" % self.motor_name)
        command = "#5,7,8dkill"
        self.set_channel_value(self.PMAC_SENDCMD, command)
        time.sleep(0.5)

        # Wait for stability
        logging.getLogger("HWR").info(
            "%s: Wait for air stability again" % self.motor_name
        )
        while self.get_channel_value(self.MOTOR_DMOV) == 0:
            time.sleep(0.2)
            current_value = self.get_value()
            self.update_value(current_value)

        # Wait for stability
        logging.getLogger("HWR").info("%s: Movement done!" % self.motor_name)
        self.update_state(self.STATES.READY)
        for i in range(2):
            logging.getLogger("user_level_log").info("Detector distance is set.")
        return value

    def _set_value(self, value):
        """Override method."""
        # As this motor moves in a special way, we delegate its movement
        # sequence to _move.
        pass



import json
import logging
import pathlib
import shutil
from datetime import (
    datetime,
    timedelta,
)
from typing import (
    List,
    Optional,
)
from zoneinfo import ZoneInfo

import requests
from pyicat_plus.client.main import IcatClient
from pyicat_plus.client.models.session import Session as ICATSession

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractLims import AbstractLims
from mxcubecore.model.lims_session import (
    Lims,
    LimsSessionManager,
    SampleInformation,
    SampleSheet,
    Session,
)


class ICATLIMS(AbstractLims):
    """
    ICAT+ client.
    """

    def __init__(self, name):
        super().__init__(name)
        HardwareObject.__init__(self, name)
        self.investigations = None
        self.icatClient = None
        self.lims_rest = None
        self.ingesters = None

    def init(self):
        self.url = self.get_property("ws_root")
        self.ingesters = self.get_property("queue_urls")
        self.investigations = []

        # Initialize ICAT client
        self.icatClient = IcatClient(
            icatplus_restricted_url=self.url,
            metadata_urls=["bcu-mq-01:61613"],
            reschedule_investigation_urls=["bcu-mq-01:61613"],
        )

    def get_lims_name(self) -> List[Lims]:
        return [
            Lims(name="DRAC", description="Data Repository for Advancing open sCience"),
        ]

    def login(
        self,
        user_name: str,
        password: str,
        session_manager: Optional[LimsSessionManager],
    ) -> LimsSessionManager:
        logging.getLogger("HWR").debug("[ICAT] authenticate %s" % (user_name))

        self.icat_session: ICATSession = self.icatClient.do_log_in(password)

        if self.icatClient is None or self.icatClient is None:
            logging.getLogger("HWR").error(
                "[ICAT] Error initializing icatClient. icatClient=%s" % (self.url)
            )
            raise RuntimeError("Could not initialize icatClient")

        # Connected to metadata icatClient
        logging.getLogger("HWR").debug(
            "[ICAT] Connected succesfully to icatClient. fullName=%s url=%s"
            % (self.icat_session["fullName"], self.url)
        )

        # Retrieving user's investigations
        sessions = self.to_sessions(self.__get_all_investigations())

        if len(sessions) == 0:
            raise Exception("No sessions available for user %s" % (user_name))

        logging.getLogger("HWR").debug(
            "[ICAT] Successfully retrieved %s sessions" % (len(sessions))
        )

        # This is done because ICATLims can be used standalone or from ESRFLims
        if session_manager is not None:
            self.session_manager = session_manager

        # Check if there is currently a session in use and if user have
        # access to that session
        if self.session_manager.active_session:
            session_found = False

            for session in sessions:
                if session.session_id == self.session_manager.active_session.session_id:
                    session_found = True
                    break

            if not session_found:
                raise Exception(
                    "Current session in-use (with id %s) not avaialble to user %s"
                    % (self.session_manager.active_session.session_id, user_name)
                )
        return self.session_manager, self.icat_session["name"], sessions

    def is_user_login_type(self) -> bool:
        return True

    def get_proposals_by_user(self, user_name):
        logging.getLogger("HWR").debug("get_proposals_by_user %s" % user_name)

        logging.getLogger("HWR").debug(
            "[ICATCLient] Read %s investigations" % len(self.lims_rest.investigations)
        )
        return self.lims_rest.to_sessions(self.lims_rest.investigations)

    def _get_loaded_pucks(self):
        """
        Retrieves all pucks from the parcels that have a defined 'sampleChangerLocation'.

        A puck is considered "loaded" if it contains the key 'sampleChangerLocation'.
        Iterates through all parcels and collects such pucks.

        Returns:
            list: A list of pucks (dicts) that have 'sampleChangerLocation' defined.
        """
        loaded_pucks = []

        if self.parcels:
            for parcel in self.parcels:
                pucks = parcel.get("content", [])
                for puck in pucks:
                    if "sampleChangerLocation" in puck:
                        # Add information about the parcel
                        puck["parcelName"] = parcel.get("name")
                        puck["parcelId"] = parcel.get("id")
                        loaded_pucks.append(puck)

        return loaded_pucks

    def get_samples(self, lims_name):
        """
        Retrieves and processes sample information from LIMS based on the provided name.

        This method:
        - Retrieves parcel data (containers like UniPucks or SpinePucks).
        - Retrieves sample sheet data.
        - Identifies and processes only loaded pucks (those with a 'sampleChangerLocation').
        - Converts each sample in the pucks into internal queue samples using `__to_sample`.

        Args:
            lims_name (str): The LIMS name or identifier used to fetch sample-related data.

        Returns:
            list: A list of processed sample objects ready for queuing.
        """
        logger = logging.getLogger("HWR")
        queue_samples = []

        try:
            session = self.session_manager.active_session
            logger.debug(
                "[ICATClient] get_samples: session_id=%s, proposal_name=%s",
                session.session_id,
                session.proposal_name,
            )

            # Load parcels (pucks)
            self.parcels = self.get_parcels()

            # Load sample sheets
            self.sample_sheets = self.get_samples_sheets()
            logger.debug(
                "[ICATClient] %d sample sheets retrieved", len(self.sample_sheets)
            )

            # Filter for loaded pucks
            self.loaded_pucks = self._get_loaded_pucks()
            logger.debug("[ICATClient] %d loaded pucks found", len(self.loaded_pucks))

            # Extract and process samples from loaded pucks
            for puck in self.loaded_pucks:
                tracking_samples = puck.get("content", [])
                puck_name = puck.get("name", "Unnamed")
                location = puck.get("sampleChangerLocation", "Unknown")

                logger.debug(
                    "[ICATClient] Found puck '%s' at position '%s' containing %d samples",
                    puck_name,
                    location,
                    len(tracking_samples),
                )

                for tracking_sample in tracking_samples:
                    sample = self.__to_sample(tracking_sample, puck, self.sample_sheets)
                    queue_samples.append(sample)

        except Exception as e:
            logger.error("[ICATClient] Error retrieving samples: %s", str(e))
            return []

        logger.debug("[ICATClient] Total %d samples read", len(queue_samples))
        self.samples = queue_samples
        return queue_samples

    def find(self, arr, atribute_name):
        for x in arr:
            if x["key"] == atribute_name:
                return x["value"]
        return ""

    def get_sample_sheet_by_id(
        self, samples: List[SampleSheet], sample_id: int
    ) -> Optional[SampleSheet]:
        """
        Retrieves a sample sheet by its unique ID.

        Args:
            samples (List[SampleSheet]): A list of Sample objects.
            sample_id (int): The unique identifier of the sample sheet to retrieve.

        Returns:
            Optional[Sample]: The Sample object if found, otherwise None.
        """
        return next((sample for sample in samples if sample.id == sample_id), None)

    def objectid_to_int(self, oid_str):
        return int(oid_str, 16)

    def int_to_objectid(self, i):
        return hex(i)[2:].zfill(24)

    def __to_sample(
        self, tracking_sample: dict, puck: dict, sample_sheets: List[SampleSheet]
    ) -> dict:
        """
        Converts a tracking sample and associated metadata into the internal sample data structure.

        This method:
        - Extracts relevant sample metadata.
        - Resolves protein acronym from the sample sheet if available.
        - Maps experiment plan details into a diffraction plan dictionary.
        - Assembles all relevant fields into a final structured sample dictionary.

        Args:
            tracking_sample (dict): The raw sample data from tracking.
            puck (dict): The puck (container) metadata associated with the sample.
            sample_sheets (List[SampleSheet]): List of sample sheets used for lookup.

        Returns:
            dict: A dictionary representing the standardized internal sample format.
        """
        # Basic identifiers
        sample_name = str(tracking_sample.get("name"))

        # MXCuBE needs to be an integer while in DRAC is a ObjectId
        # Mongo @BES needs to be smaller then 8 bytes
        sample_id = int(str(self.objectid_to_int(tracking_sample.get("id")))[0:6])
        # id to the sample sheet declared in the user portal
        sample_sheet_id = tracking_sample.get("sampleId")
        # identifier that points to the sample tracking
        trackingSampleId = tracking_sample.get("_id")

        logging.getLogger("HWR").debug(
            "[ICATClient] Sample ids sample_id=%s sample_sheet_id=%s trackingSampleId=%s",
            sample_id,
            sample_sheet_id,
            trackingSampleId,
        )

        sample_location = tracking_sample.get("sampleContainerPosition")
        puck_location = str(puck.get("sampleChangerLocation", "Unknown"))
        puck_name = puck.get("name", "UnknownPuck")
        parcel_name = puck.get("parcelName")
        parcel_id = puck.get("parcelId")
        # Determine protein acronym using sample sheet if available
        protein_acronym = sample_name  # Default fallback

        sample_sheet = self.get_sample_sheet_by_id(sample_sheets, sample_sheet_id)
        if sample_sheet:
            protein_acronym = sample_sheet.name

        experiment_plan = tracking_sample.get("experimentPlan", {})

        processing_plan = tracking_sample.get("processingPlan", [])
        search_models = None
        reference = None

        if processing_plan:
            try:
                parsed_plan = json.loads(processing_plan[0]["value"])[0]
            except (ValueError, KeyError, IndexError, TypeError) as e:
                parsed_plan = {}

            try:
                search_models = parsed_plan.get("search_models")
            except Exception:
                search_models = None

            try:
                reference = parsed_plan.get("reference")
            except Exception:
                reference = None

        comments = tracking_sample.get("comments")

        return {
            "sampleName": sample_name,
            "sampleId": sample_id,
            "sample_sheet_id": sample_sheet_id,
            "trackingSampleId": trackingSampleId,
            "proteinAcronym": protein_acronym,
            "searchModels": search_models,
            "reference": reference,
            "sampleLocation": sample_location,
            "containerCode": puck_name,
            "containerSampleChangerLocation": puck_location,
            "SampleTrackingParcel_name": parcel_name,
            "SampleTrackingParcel_id": parcel_id,
            "SampleTrackingContainer_id": puck_name,
            "SampleTrackingContainer_name": parcel_id,
            "smiles": None,  # Placeholder for future chemical structure info
            "experimentType": self.find(experiment_plan, "workflowType"),
            "crystalSpaceGroup": self.find(experiment_plan, "forceSpaceGroup"),
            "diffractionPlan": {
                # "diffractionPlanId": 457980, TODO: do we need this?
                "experimentKind": self.find(experiment_plan, "experimentKind"),
                "numberOfPositions": self.find(experiment_plan, "numberOfPositions"),
                "observedResolution": self.find(experiment_plan, "observedResolution"),
                "preferredBeamDiameter": self.find(
                    experiment_plan, "preferredBeamDiameter"
                ),
                "radiationSensitivity": self.find(
                    experiment_plan, "radiationSensitivity"
                ),
                "requiredCompleteness": self.find(
                    experiment_plan, "requiredCompleteness"
                ),
                "requiredMultiplicity": self.find(
                    experiment_plan, "requiredMultiplicity"
                ),
                "requiredResolution": self.find(experiment_plan, "requiredResolution"),
            },
            "cellA": self.find(experiment_plan, "unit_cell_a"),
            "cellB": self.find(experiment_plan, "unit_cell_b"),
            "cellC": self.find(experiment_plan, "unit_cell_c"),
            "cellAlpha": self.find(experiment_plan, "unit_cell_alpha"),
            "cellBeta": self.find(experiment_plan, "unit_cell_beta"),
            "cellGamma": self.find(experiment_plan, "unit_cell_gamma"),
            "experimentPlan": experiment_plan,
            "processingPlan": processing_plan,
            "comments": comments,
        }

    def create_session(self, session_dict):
        pass

    def _store_data_collection_group(self, group_data):
        pass

    def store_robot_action(self, proposal_id: str):
        raise Exception("Not implemented")

    @property
    def only_staff_session_selection(self):
        return bool(self.get_property("only_staff_session_selection", default=True))

    @property
    def filter(self):
        return self.get_property("filter", None)

    @property
    def override_beamline_name(self):
        return self.get_property(
            "override_beamline_name", HWR.beamline.session.beamline_name
        )

    @property
    def compatible_beamlines(self):
        return self.get_property(
            "compatible_beamlines", HWR.beamline.session.beamline_name
        )

    @property
    def data_portal_url(self):
        return self.get_property("data_portal_url", None)

    @property
    def user_portal_url(self):
        return self.get_property("user_portal_url", None)

    @property
    def logbook_url(self):
        return self.get_property("logbook_url", None)

    @property
    def before_offset_days(self):
        return self.get_property("before_offset_days", "1")

    @property
    def after_offset_days(self):
        return self.get_property("after_offset_days", "1")

    def _string_to_format_date(self, date: str, format: str) -> str:
        if date is not None:
            date_time = self._tz_aware_fromisoformat(date)
            if date_time is not None:
                return date_time.strftime(format)
        return ""

    def _string_to_date(self, date: str) -> str:
        return self._string_to_format_date(date, "%Y%m%d")

    def _string_to_time(self, date: str) -> str:
        return self._string_to_format_date(date, "%H:%M:%S")

    def _tz_aware_fromisoformat(self, date: str) -> datetime:
        try:
            return datetime.fromisoformat(date).astimezone()
        except Exception:
            return None

    def set_active_session_by_id(self, session_id: str) -> Session:
        logging.getLogger("HWR").debug(f"set_active_session_by_id: {session_id}")

        if self.is_session_already_active(self.session_manager.active_session):
            return self.session_manager.active_session

        sessions = self.session_manager.sessions

        logging.getLogger("HWR").debug(f"Sessions: {len(sessions)}")

        if len(sessions) == 0:
            logging.getLogger("HWR").error(
                "Session list is empty. No session candidates"
            )
            raise Exception("No sessions available")

        if len(sessions) == 1:
            self.session_manager.active_session = sessions[0]
            logging.getLogger("HWR").debug(
                "Session list contains a single session. proposal_name=%s",
                self.session_manager.active_session.proposal_name,
            )
            return self.session_manager.active_session

        session_list = [obj for obj in sessions if obj.session_id == session_id]
        if len(session_list) != 1:
            raise Exception(
                "Session not found in the local list of sessions. session_id="
                + session_id
            )
        self.session_manager.active_session = session_list[0]
        return self.session_manager.active_session

    def allow_session(self, session: Session):
        self.active_session = session
        logging.getLogger("HWR").debug(
            "[ICAT] allow_session investigationId=%s", session.session_id
        )
        self.icatClient.reschedule_investigation(session.session_id)

    def get_session_by_id(self, id: str):
        logging.getLogger("HWR").debug(
            "[ICAT] get_session_by_id investigationId=%s investigations=%s",
            id,
            str(len(self.investigations)),
        )
        investigation_list = list(filter(lambda p: p["id"] == id, self.investigations))
        if len(investigation_list) == 1:
            self.investigation = investigation_list[0]
            return self.__to_session(investigation_list[0])
        logging.getLogger("HWR").warn(
            "[ICAT] No investigation found. get_session_by_id investigationId=%s investigations=%s",
            id,
            str(len(self.investigations)),
        )
        return None

    def __get_all_investigations(self):
        """Returns all investigations by user. An investigation corresponds to
        one experimental session. It returns an empty array in case of error"""
        try:
            self.investigations = []
            logging.getLogger("HWR").debug(
                "[ICAT] __get_all_investigations before=%s after=%s beamline=%s isInstrumentScientist=%s isAdministrator=%s compatible_beamlines=%s"
                % (
                    self.before_offset_days,
                    self.after_offset_days,
                    self.override_beamline_name,
                    self.icat_session["isInstrumentScientist"],
                    self.icat_session["isAdministrator"],
                    self.compatible_beamlines,
                )
            )

            if self.icat_session is not None and (
                self.icat_session["isAdministrator"]
                or self.icat_session["isInstrumentScientist"]
            ):
                # Setting up of the session done by admin or staff
                self.investigations = self.icatClient.get_investigations_by(
                    start_date=datetime.today()
                    - timedelta(days=float(self.before_offset_days)),
                    end_date=datetime.today()
                    + timedelta(days=float(self.after_offset_days)),
                    instrument_name=self.compatible_beamlines,
                )
            elif self.only_staff_session_selection:
                if self.session_manager.active_session is None:
                    # If no session selected and only staff is allowed then print warning an return no investigations
                    logging.getLogger("HWR").warning(
                        "No session selected. Only staff can select a session"
                    )
                    return []

                self.investigations = self.icatClient.get_investigations_by(
                    ids=[self.session_manager.active_session.session_id]
                )
            else:
                self.investigations = self.icatClient.get_investigations_by(
                    filter=self.filter,
                    instrument_name=self.compatible_beamlines,
                    start_date=datetime.today()
                    - timedelta(days=float(self.before_offset_days)),
                    end_date=datetime.today()
                    + timedelta(days=float(self.after_offset_days)),
                )
            logging.getLogger("HWR").debug(
                "[ICAT] __get_all_investigations retrieved %s investigations"
                % len(self.investigations)
            )
            return self.investigations
        except Exception as e:
            self.investigations = []
            logging.getLogger("HWR").error("[ICAT] __get_all_investigations %s " % e)
        return self.investigations

    def __get_proposal_number_by_investigation(self, investigation):
        """
        Given an investigation it returns the proposal number.
        Example: investigation["name"] = "MX-1234"
        returns: 1234

        TODO: this might not work for all type of proposals (example: TEST proposals)
        """
        return (
            investigation["name"]
            .replace(investigation["type"]["name"], "")
            .replace("-", "")
        )

    def _get_data_portal_url(self, investigation):
        try:
            return (
                self.data_portal_url.replace("{id}", str(investigation["id"]))
                if self.data_portal_url is not None
                else ""
            )
        except Exception:
            return ""

    def _get_logbook_url(self, investigation):
        try:
            return (
                self.logbook_url.replace("{id}", str(investigation["id"]))
                if self.logbook_url is not None
                else ""
            )
        except Exception:
            return ""

    def _get_user_portal_url(self, investigation):
        try:
            return (
                self.user_portal_url.replace(
                    "{id}", str(investigation["parameters"]["Id"])
                )
                if self.user_portal_url is not None
                and investigation["parameters"]["Id"] is not None
                else ""
            )
        except Exception:
            return ""

    def __get_investigation_parameter_by_name(
        self, investigation: dict, parameter_name: str
    ) -> str:
        """
        Gets the metadata of the parameters in an investigation
        Returns the value of the specified parameter if it exists,
        otherwise returns an empty string.
        """
        return investigation.get("parameters", {}).get(parameter_name, None)

    def __to_session(self, investigation) -> Session:
        """This methods converts a ICAT investigation into a session"""

        actual_start_date = (
            investigation["parameters"]["__actualStartDate"]
            if "__actualStartDate" in investigation["parameters"]
            else investigation["startDate"]
        )
        actual_end_date = (
            investigation["parameters"]["__actualEndDate"]
            if "__actualEndDate" in investigation["parameters"]
            else investigation.get("endDate", None)
        )

        instrument_name = investigation["instrument"]["name"]

        # If session has been rescheduled new date is overwritten
        return Session(
            code=investigation["type"]["name"],
            number=self.__get_proposal_number_by_investigation(investigation),
            title=f"{investigation['title']}",
            session_id=investigation["id"],
            proposal_id=investigation["id"],
            proposal_name=investigation["name"],
            beamline_name=instrument_name,
            comments="",
            start_datetime=investigation.get(
                "startDate", None
            ),  # self._string_to_date(investigation.get("startDate", None)),
            start_date=self._string_to_date(investigation.get("startDate", None)),
            start_time=self._string_to_time(investigation.get("startDate", None)),
            end_datetime=investigation.get("endDate", None),
            end_date=self._string_to_date(
                investigation.get("endDate", None)
            ),  # self._string_to_time(investigation.get("endDate", None)),
            end_time=self._string_to_time(investigation.get("endDate", None)),
            actual_start_date=self._string_to_date(actual_start_date),
            actual_start_time=self._string_to_time(actual_start_date),
            actual_end_date=self._string_to_date(actual_end_date),
            actual_end_time=self._string_to_time(actual_end_date),
            nb_shifts=3,
            scheduled=self.is_scheduled_on_host_beamline(instrument_name),
            is_scheduled_time=self.is_scheduled_now(actual_start_date, actual_end_date),
            is_scheduled_beamline=self.is_scheduled_on_host_beamline(instrument_name),
            data_portal_URL=self._get_data_portal_url(investigation),
            user_portal_URL=self._get_user_portal_url(investigation),
            logbook_URL=self._get_logbook_url(investigation),
            is_rescheduled=(
                True if "__actualEndDate" in investigation["parameters"] else False
            ),
            volume=self.__get_investigation_parameter_by_name(
                investigation, "__volume"
            ),
            sample_count=self.__get_investigation_parameter_by_name(
                investigation, "__sampleCount"
            ),
            dataset_count=self.__get_investigation_parameter_by_name(
                investigation, "__datasetCount"
            ),
        )

    def get_full_user_name(self):
        return self.icat_session["fullName"]

    def get_user_name(self):
        return self.icat_session["username"]

    def to_sessions(self, investigations):
        return [self.__to_session(investigation) for investigation in investigations]

    def get_parcels(self):
        """Returns the parcels associated to an investigation"""
        try:
            logging.getLogger("HWR").debug(
                "[ICAT] Retrieving parcels by investigation_id %s "
                % (self.session_manager.active_session.session_id)
            )
            parcels = self.icatClient.get_parcels_by(
                self.session_manager.active_session.session_id
            )
            logging.getLogger("HWR").debug(
                "[ICAT] Successfully retrieved %s parcels" % (len(parcels))
            )
            return parcels
        except Exception as e:
            logging.getLogger("HWR").error(
                "[ICAT] get_parcels_by_investigation_id %s " % (str(e))
            )
        return []

    def get_samples_sheets(self) -> List[SampleSheet]:
        """Returns the samples sheets associated to an investigation"""
        try:
            logging.getLogger("HWR").debug(
                "[ICAT] Retrieving samples by investigation_id %s "
                % (self.session_manager.active_session.session_id)
            )
            samples = self.icatClient.get_samples_by(
                self.session_manager.active_session.session_id
            )
            logging.getLogger("HWR").debug(
                "[ICAT] Successfully retrieved %s samples" % (len(samples))
            )
            # Convert to object
            return [SampleSheet.parse_obj(sample) for sample in samples]
        except Exception as e:
            logging.getLogger("HWR").error(
                "[ICAT] get_samples_by_investigation_id %s " % (str(e))
            )
        return []

    def echo(self):
        """Mockup for the echo method."""
        return True

    def is_connected(self):
        return self.login_ok

    def add_beamline_configuration_metadata(self, metadata, beamline_config):
        """
        This is the mapping betweeh the beamline_config dict and the ICAt keys
        in case they exist then they will be added to the metadata of the dataset
        """
        if beamline_config is not None:
            key_mapping = {
                "detector_px": "InstrumentDetector01_beam_center_x",
                "detector_py": "InstrumentDetector01_beam_center_y",
                "beam_divergence_vertical": "InstrumentBeam_vertical_incident_beam_divergence",
                "beam_divergence_horizontal": "InstrumentBeam_horizontal_incident_beam_divergence",
                "polarisation": "InstrumentBeam_final_polarization",
                "detector_model": "InstrumentDetector01_model",
                "detector_manufacturer": "InstrumentDetector01_manufacturer",
                "synchrotron_name": "InstrumentSource_name",
                "monochromator_type": "InstrumentMonochromatorCrystal_type",
                "InstrumentDetector01_type": "detector_type",
            }

            for config_key, metadata_key in key_mapping.items():
                if hasattr(beamline_config, config_key):
                    metadata[metadata_key] = getattr(beamline_config, config_key)

    def find_sample_by_sample_id(self, sample_id):
        return next(
            (sample for sample in self.samples if sample["limsID"] == sample_id),
            None,
        )

    def _get_sample_position(self):
        """
        Returns the position of the puck in the samples changer and the position f the sample within the puck
        """
        try:
            queue_entry = HWR.beamline.queue_manager.get_current_entry()
            sample_node = queue_entry.get_data_model().get_sample_node()
            location = sample_node.location  # Example: (8,2,5)

            if len(location) == 3:
                (cell, puck, sample_position) = location
            else:
                cell = 1
                (puck, sample_position) = location

            position = None
            try:
                if cell is not None and puck is not None:
                    position = int(cell * 3) + int(puck)
            except Exception as e:
                logging.getLogger("HWR").exception(e)
            return position, sample_position
        except Exception as e:
            logging.getLogger("HWR").exception(e)

    def store_beamline_setup(self, session_id: str, bl_config_dict: dict):
        pass

    def store_image(self, image_dict: dict):
        pass

    def store_energy_scan(self, energyscan_dict: dict):
        pass

    def store_xfe_spectrum(self, xfespectrum_dict: dict):
        pass

    def store_workflow(self, workflow_dict: dict):
        pass

    def store_data_collection(self, mx_collection, bl_config=None):
        # stores the dictionary with the information about the beamline to be sent when a dataset is produced
        self.beamline_config = bl_config

    def update_data_collection(self, mx_collection):
        pass

    def _get_oscillation_end(self, oscillation_sequence):
        return float(oscillation_sequence["start"]) + (
            float(oscillation_sequence["range"])
            - float(oscillation_sequence["overlap"])
        ) * float(oscillation_sequence["number_of_images"])

    def _get_rotation_axis(self, oscillation_sequence):
        if "kappaStart" in oscillation_sequence:
            if (
                oscillation_sequence["kappaStart"] != 0
                and oscillation_sequence["kappaStart"] != -9999
            ):
                return "Omega"
        return "Phi"

    def __get_sample_information_by(
        self, sample_id: str
    ) -> Optional[SampleInformation]:
        """
        Fetches sample metadata and associated resources based on the sample ID.

        Parameters:
            sample_id (str): The unique identifier for the sample.

        Returns:
            Optional[SampleInformation]: Returns a SampleInformation object or None.
        """
        try:
            token = self.icat_session["sessionId"]
            url = f"{self.url}/catalogue/{token}/files?sampleId={sample_id}"
            response = requests.get(url, timeout=3)
            response.raise_for_status()  # Raise an exception for bad status codes
            return SampleInformation(
                **response.json(),
            )  # Parse the response into a SampleInformation model
        except requests.exceptions.RequestException as e:
            logging.error("Failed to fetch sample information for %s: %s", sample_id, e)
        return None

    def _download_resources(self, sample_id, resources, output_folder: str):
        """
        Downloads resources related to a given sample and saves them to the specified directory.

        Parameters:
            sample (str): Sample identifier.
            output_folder (str): Directory where files will be saved.

        Returns:
            dict: A dictionary containing the paths of the downloaded files.
        """
        downloaded_files = []
        for resource in resources:
            resource_folder = pathlib.Path(output_folder) / (
                resource.groupName if resource.groupName else ""
            )
            resource_folder.mkdir(
                parents=True,
                exist_ok=True,
            )  # Ensure the folder exists

            try:
                token = self.icat_session["sessionId"]
                url = f"{self.url}/catalogue/{token}/files/download?sampleId={sample_id}&resourceId={resource.id}"
                response = requests.get(url, stream=True, timeout=3)
                response.raise_for_status()

                file_path = resource_folder / resource.filename
                with file_path.open("wb") as file:
                    for chunk in response.iter_content(
                        chunk_size=8192,
                    ):  # Efficient chunked download
                        file.write(chunk)

                downloaded_files.append(str(file_path))
                logging.info("Downloaded %s to %s", resource.filename, file_path)

            except requests.exceptions.RequestException as e:
                logging.error("Failed to download %s: %s", resource.filename, e)

        return downloaded_files

    def finalize_data_collection(self, collection_parameters):
        logging.getLogger("HWR").info("Storing datacollection in ICAT")

        try:
            fileinfo = collection_parameters["fileinfo"]
            directory = pathlib.Path(fileinfo["directory"])
            dataset_name = directory.name
            # Determine the scan type
            if dataset_name.endswith("mesh"):
                scan_type = "mesh"
            elif dataset_name.endswith("line"):
                scan_type = "line"
            elif dataset_name.endswith("characterisation"):
                scan_type = "characterisation"
            elif dataset_name.endswith("datacollection"):
                scan_type = "datacollection"
            else:
                scan_type = collection_parameters["experiment_type"]

            workflow_params = collection_parameters.get("workflow_parameters", {})
            workflow_type = workflow_params.get("workflow_type")

            if workflow_type is None:
                if not directory.name.startswith("run"):
                    dataset_name = fileinfo["prefix"]

            try:
                dt_naive = datetime.strptime(
                    collection_parameters.get("collection_start_time"),
                    "%Y-%m-%d %H:%M:%S%z",
                )
                dt_aware = dt_naive.replace(tzinfo=ZoneInfo("Europe/Paris"))
                start_time = dt_aware.isoformat(timespec="microseconds")
                end_time = datetime.now(ZoneInfo("Europe/Paris")).isoformat()
            except RuntimeError:
                logging.getLogger("HWR").error("Failed to parse start and end time")

            if collection_parameters["sample_reference"]["acronym"]:
                sample_name = (
                    collection_parameters["sample_reference"]["acronym"]
                    + "-"
                    + collection_parameters["sample_reference"]["sample_name"]
                )
            else:
                sample_name = collection_parameters["sample_reference"][
                    "sample_name"
                ].replace(":", "-")

            logging.getLogger("HWR").info(f"LIMS sample name {sample_name}")
            oscillation_sequence = collection_parameters["oscillation_sequence"][0]

            beamline = HWR.beamline.session.beamline_name.lower()
            distance = HWR.beamline.detector.distance.get_value()
            proposal = f"{HWR.beamline.session.proposal_code}{HWR.beamline.session.proposal_number}"
            metadata = {
                "MX_beamShape": collection_parameters.get("beamShape"),
                "sampleId": collection_parameters.get("blSampleId"),
                "MX_beamSizeAtSampleX": collection_parameters.get("beamSizeAtSampleX"),
                "MX_beamSizeAtSampleY": collection_parameters.get("beamSizeAtSampleY"),
                "MX_dataCollectionId": collection_parameters.get("collection_id"),
                "MX_detectorDistance": distance,
                "MX_directory": str(directory),
                "MX_exposureTime": oscillation_sequence["exposure_time"],
                "MX_flux": collection_parameters.get("flux"),
                "MX_fluxEnd": collection_parameters.get("flux_end"),
                "MX_positionName": collection_parameters.get("position_name"),
                "MX_numberOfImages": oscillation_sequence["number_of_images"],
                "MX_oscillationRange": oscillation_sequence["range"],
                "MX_axis_start": oscillation_sequence["start"],
                "MX_oscillationOverlap": oscillation_sequence["overlap"],
                "MX_resolution": collection_parameters.get("resolution"),
                "MX_resolution_at_corner": collection_parameters.get(
                    "resolutionAtCorner"
                ),
                "scanType": scan_type,
                "MX_startImageNumber": oscillation_sequence["start_image_number"],
                "MX_template": fileinfo["template"],
                "MX_transmission": collection_parameters.get("transmission"),
                "MX_xBeam": collection_parameters.get("xBeam"),
                "MX_yBeam": collection_parameters.get("yBeam"),
                "Sample_name": sample_name,
                "InstrumentMonochromator_wavelength": collection_parameters.get(
                    "wavelength"
                ),
                "Workflow_name": workflow_params.get("workflow_name"),
                "Workflow_type": workflow_params.get("workflow_type"),
                "Workflow_id": workflow_params.get("workflow_uid"),
                "MX_kappa_settings_id": workflow_params.get(
                    "workflow_kappa_settings_id"
                ),
                "MX_characterisation_id": workflow_params.get(
                    "workflow_characterisation_id"
                ),
                "MX_position_id": workflow_params.get("workflow_position_id"),
                "group_by": workflow_params.get("workflow_group_by"),
                "startDate": start_time,
                "endDate": end_time,  # strftime("%Y-%m-%d %H:%M:%S"),
            }

            # This forces the ingester to associate the dataset to the experiment by ID
            if self.session_manager.active_session.session_id:
                metadata["investigationId"] = (
                    self.session_manager.active_session.session_id
                )

            metadata["SampleTrackingContainer_type"] = "UNIPUCK"
            metadata["SampleTrackingContainer_capacity"] = "16"
            (position, sample_position) = self._get_sample_position()
            metadata["SampleChanger_position"] = position
            metadata["SampleTrackingContainer_position"] = sample_position
            # Find sample by sampleId
            sample = HWR.beamline.lims.find_sample_by_sample_id(
                collection_parameters.get("blSampleId")
            )

            if sample is not None:
                metadata["SampleProtein_acronym"] = sample.get("proteinAcronym")
                metadata["SampleTrackingContainer_id"] = sample.get(
                    "containerCode"
                )  # containerCode instead of sampletrackingcontainer_id for ISPyB's compatibility
                metadata["SampleTrackingParcel_id"] = sample.get(
                    "SampleTrackingParcel_id"
                )
                metadata["SampleTrackingParcel_name"] = sample.get(
                    "SampleTrackingParcel_name"
                )

                reference_paths = None
                search_models_paths = None
                try:
                    if (
                        # self.download_sample_resources and
                        sample.get("sample_sheet_id") is not None
                        and scan_type == "datacollection"
                    ):
                        sample_info = self.__get_sample_information_by(
                            sample.get("sample_sheet_id")
                        )

                        sample_resource_folder = directory

                        if sample.get("reference") is not None:
                            # Assuming `sample_info` is your SampleInformation instance
                            matching_resources = [
                                res
                                for res in sample_info.resources
                                if res.filename == sample.get("reference")
                            ]
                            reference_paths = self._download_resources(
                                sample.get("sample_sheet_id"),
                                matching_resources,
                                sample_resource_folder,
                            )

                        if sample.get("searchModels") is not None:
                            matching_resources = [
                                res
                                for res in sample_info.resources
                                if res.groupName == sample.get("searchModels")
                            ]
                            search_models_paths = self._download_resources(
                                sample.get("sample_sheet_id"),
                                matching_resources,
                                sample_resource_folder,
                            )

                except RuntimeError:
                    logging.getLogger("HWR").error("Failed to get download resources")

            self.add_beamline_configuration_metadata(metadata, self.beamline_config)

            # MX_axis_end
            try:
                metadata["MX_axis_end"] = self._get_oscillation_end(
                    oscillation_sequence
                )
            except RuntimeError:
                logging.getLogger("HWR").error("Failed to get MX_axis_end")

            # MX_axis_end
            try:
                metadata["MX_axis_range"] = self._get_rotation_axis(
                    oscillation_sequence
                )
            except RuntimeError:
                logging.getLogger("HWR").exception("Failed to get MX_axis_end")

            icat_metadata_path = pathlib.Path(directory) / "metadata.json"
            with open(icat_metadata_path, "w") as f:
                # We add the processing and experiment plan only in the metadata.json
                # it will not work thought pyicat-plus
                merged = metadata.copy()
                try:
                    if sample is not None:
                        merged["experimentPlan"] = sample.get("experimentPlan")
                        merged["processingPlan"] = sample.get("processingPlan")
                        merged["search_models_paths"] = search_models_paths
                        merged["reference"] = reference_paths
                except Exception:
                    logging.getLogger("HWR").exception(
                        "Failed to get merged sample plan"
                    )

                f.write(json.dumps(merged, indent=4))
            # Create ICAT gallery
            gallery_path = directory / "gallery"
            gallery_path.mkdir(mode=0o755, exist_ok=True)
            for snapshot_index in range(1, 5):
                key = f"xtalSnapshotFullPath{snapshot_index}"
                if key in collection_parameters:
                    snapshot_path = pathlib.Path(collection_parameters[key])
                    if snapshot_path.exists():
                        logging.getLogger("HWR").debug(
                            f"Copying snapshot index {snapshot_index} to gallery"
                        )
                        shutil.copy(snapshot_path, gallery_path)

            try:
                beamline = self._get_scheduled_beamline()
                logging.getLogger("HWR").info(
                    f"Dataset Beamline={beamline} Current Beamline={HWR.beamline.session.beamline_name}"
                )
            except Exception:
                logging.getLogger("HWR").exception(
                    "Failed to get _get_scheduled_beamline"
                )

            # __actualInstrument is a dataset parameter that indicates where the dataset has been actually collected
            # only filled when it does not match the scheduled beamline
            try:
                if (
                    self.active_session is None
                    or not self.active_session.is_scheduled_beamline
                ):
                    metadata["__actualInstrument"] = HWR.beamline.session.beamline_name
            except Exception:
                logging.getLogger("HWR").exception("")

            self.icatClient.store_dataset(
                beamline=beamline,
                proposal=proposal,
                dataset=dataset_name,
                path=str(directory),
                metadata=metadata,
            )
            logging.getLogger("HWR").debug("Done uploading to ICAT")
        except Exception as e:
            logging.getLogger("HWR").exception("Failed uploading to ICAT")

    def _get_scheduled_beamline(self):
        """
        This returns the beamline where the session has been scheduled (in case of a different beamline)
        otherwise it returns the name of the beamline as set in the properties
        """
        active_session = self.session_manager.active_session

        if active_session is None or active_session.is_scheduled_beamline:
            return HWR.beamline.session.beamline_name.lower()

        beamline = str(active_session.beamline_name.lower())
        logging.getLogger("HWR").info(
            f"Session have been moved to another beamline: {beamline}"
        )
        return beamline

    def update_bl_sample(self, bl_sample: str):
        """
        Creates or stos a BLSample entry.
        # NBNB update doc string
        :param sample_dict: A dictionary with the properties for the entry.
        :type sample_dict: dict
        """
        pass



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
[Name]
GenericVideo

[Description]
This module declares class GenericVideo.

This class is not meant to be instanced directly but as
the base class for classes providing access to Video in MXCuBE
"""

import abc
import logging
import os
import sys
import time
import warnings
from io import BytesIO

import gevent
import numpy as np

try:
    import cv2
except ImportError:
    pass

from mxcubecore.BaseHardwareObjects import HardwareObject

module_names = ["qt", "PyQt5", "PyQt4"]

if any(mod in sys.modules for mod in module_names):
    USEQT = True
    from mxcubecore.utils.qt_import import (
        QImage,
        QPixmap,
        QSize,
    )
else:
    USEQT = False
    from PIL import Image


class AbstractVideoDevice(HardwareObject):
    default_cam_encoding = "yuv422p"
    default_poll_interval = 50
    default_cam_type = "basler"
    default_scale_factor = 1.0

    def __init__(self, name):
        super().__init__(name)

        self.cam_mirror = None
        self.cam_encoding = None
        self.cam_gain = None
        self.cam_exposure = None
        self.poll_interval = None
        self.cam_type = None
        self.cam_scale_factor = None
        self.cam_name = None

        self.raw_image_dimensions = [None, None]
        self.image_dimensions = [None, None]
        self.image_polling = None
        self.image_format = None  # not used
        self.default_cam_encoding = None
        self.default_poll_interval = None

        self.decoder = None
        self.scale = None

    def init(self):
        """Initialise the values from config and set default values,
        when appropriate
        """
        self.cam_name = self.get_property("name", "camera")

        try:
            self.cam_mirror = eval(self.get_property("mirror"))
        except TypeError:
            self.cam_mirror = [False, False]

        try:
            self.cam_encoding = self.get_property("encoding").lower()
        except AttributeError:
            # apply default value
            self.cam_encoding = AbstractVideoDevice.default_cam_encoding

        scale = self.get_property("scale")
        try:
            self.cam_scale_factor = eval(scale)
        except TypeError:
            logging.getLogger().warning(
                "%s: failed to interpret scale factor for camera.\nUsing default.",
                self.id,
            )
            self.cam_scale_factor = self.default_scale_factor

        self.poll_interval = self.get_property("interval", self.default_poll_interval)

        try:
            self.cam_gain = float(self.get_property("gain"))
        except TypeError:
            pass

        try:
            self.cam_exposure = float(self.get_property("exposure"))
        except TypeError:
            # apply default value
            self.cam_exposure = self.poll_interval / 1000.0

        self.scale = self.get_property("scale", 1.0)

        try:
            self.cam_type = self.get_property("type").lower()
        except AttributeError:
            # apply default value
            self.cam_type = self.default_cam_type

        # Apply values
        self.set_video_live(False)
        time.sleep(0.1)
        self.set_cam_encoding(self.cam_encoding)
        self.set_exposure_time(self.cam_exposure)

        if self.cam_gain is not None:
            self.set_gain(self.cam_gain)

        self.image_dimensions = self.get_image_dimensions()

        # Start polling greenlet
        if self.image_polling is None:
            self.set_video_live(True)
            self.change_owner()

            logging.getLogger("HWR").info("Starting polling for camera")
            self.image_polling = gevent.spawn(
                self.do_image_polling, self.poll_interval / 1000.0
            )
            self.image_polling.link_exception(self.polling_ended_exc)
            self.image_polling.link(self.polling_ended)

    def get_camera_name(self):
        """Get the camera name.
        Returns:
            (str): The name
        """
        return self.cam_name

    def polling_ended(self, gl=None):
        logging.getLogger("HWR").info("Polling ended for qt camera")

    def polling_ended_exc(self, gl=None):
        logging.getLogger("HWR").info("Polling ended exception for qt camera")

    # -------- Generic methods --------

    def get_new_image(self):
        """
        Descript. :
        """
        raw_buffer, width, height = self.get_image()

        if raw_buffer is not None and raw_buffer.any():
            if self.decoder:
                raw_buffer = self.decoder(raw_buffer)
                qimage = QImage(
                    raw_buffer, width, height, width * 3, QImage.Format_RGB888
                )
            else:
                qimage = QImage(raw_buffer, width, height, QImage.Format_RGB888)

            if self.cam_mirror is not None:
                qimage = qimage.mirrored(self.cam_mirror[0], self.cam_mirror[1])

            if self.scale != 1:
                dims = self.get_image_dimensions()  # should be already scaled
                qimage = qimage.scaled(QSize(dims[0], dims[1]))

            qpixmap = QPixmap(qimage)
            self.emit("imageReceived", qpixmap)
            return qimage.copy()

    def get_jpg_image(self):
        """Reads`raw_data` image `[1D numpy array of np.uint16]` from
        `self.get_image()` and convert it to .jpg image.
        For now this function allows to deal with any RGB encoded
        video data. Emit imageReceived signal with the jpeg image.

        Returns:
            (bytes): Converted to jpeg image.
        """
        raw_buffer, width, height = self.get_image()

        if raw_buffer is not None and raw_buffer.any():
            image = Image.frombytes("RGB", (width, height), raw_buffer)
            buffer = BytesIO()
            image.save(buffer, "JPEG")
            jpg_img = buffer.getvalue()
            if jpg_img is not None:
                self.emit("imageReceived", jpg_img, width, height)
            return jpg_img
        return None

    def get_cam_type(self):
        """Get the camera type
        Returns:
            (): Camera type.
        """
        return self.cam_type

    def y8_2_rgb(self, raw_buffer):
        """Convert Y8 to RGB.
        Args:
            raw_buffer: Image
        Returns:
            (): Converted image.
        """
        image = np.fromstring(raw_buffer, dtype=np.uint8)
        raw_dims = self.get_raw_image_size()
        image.resize(raw_dims[1], raw_dims[0], 1)
        return cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)

    def y16_2_rgb(self, raw_buffer):
        """Convert Y16 to RGB.
        Args:
            raw_buffer: Image
        Returns:
            (): Converted image.
        """
        image = np.fromstring(raw_buffer, dtype=np.uint8)
        raw_dims = self.get_raw_image_size()
        np.resize(image, (raw_dims[1], raw_dims[0], 2))
        return cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)

    def yuv_2_rgb(self, raw_buffer):
        """Convert YUV to RGB.
        Args:
            raw_buffer: Image
        Returns:
            (): Converted image.
        """
        image = np.fromstring(raw_buffer, dtype=np.uint8)
        raw_dims = self.get_raw_image_size()
        image.resize(raw_dims[1], raw_dims[0], 2)
        return cv2.cvtColor(image, cv2.COLOR_YUV2RGB_UYVY)

    def bayer_rg16_2_rgb(self, raw_buffer):
        """Convert BAYER RG16 to RGB.
        Args:
            raw_buffer: Image
        Returns:
            (): Converted image.
        """
        image = np.fromstring(raw_buffer, dtype=np.uint16)
        raw_dims = self.get_raw_image_size()
        image.resize(raw_dims[1], raw_dims[0])
        out_buffer = cv2.cvtColor(image, cv2.COLOR_BayerRG2BGR)
        if out_buffer.ndim == 3 and out_buffer.itemsize > 1:
            # decoding bayer16 gives 12 bit values => scale to 8 bit
            out_buffer = np.right_shift(out_buffer, 4).astype(np.uint8)
        return out_buffer

    def save_snapshot(self, filename, image_type="PNG"):
        """Save snapshot image"""
        if USEQT:
            qimage = self.get_new_image()
            qimage.save(filename, image_type)
        else:
            jpgstr = self.get_jpg_image()
            open(filename, "w").write(jpgstr)

    def take_snapshot(self, bw=None, return_as_array=True):
        """Take the snapshot.
        Args:
            bs(bool): Return grayscale image (True)
            return_as_array(bool): Return the image as array. Default is True.
        Returns:
            (): Snapshot image.
        """
        self.get_snapshot(bw, return_as_array)

    def get_snapshot(self, bw=None, return_as_array=True):
        """Get the snapshot.
        Args:
            bs(bool): Return grayscale image (True)
            return_as_array(bool): Return the image as array. Default is True.
        Returns:
            (): Snapshot image.
        """
        warnings.warn(
            "Deprecated method, call take_snapshot instead", DeprecationWarning
        )
        if not USEQT:
            print("get snapshot not implemented yet for non-qt mode")
            return None

        qimage = self.get_new_image()
        if return_as_array:
            qimage = qimage.convertToFormat(4)
            ptr = qimage.bits()
            ptr.setsize(qimage.byteCount())
            image_array = np.array(ptr).reshape(qimage.height(), qimage.width(), 4)
            if bw:
                return np.dot(image_array[..., :3], [0.299, 0.587, 0.144])
            return image_array

        if bw:
            return qimage.convertToFormat(QImage.Format_Mono)
        return qimage

    def get_scaling_factor(self):
        """Get the Scaling factor.
        Returns:
            (float): Scaling factor or None if does not exists
        """
        return self.cam_scale_factor

    get_image_zoom = get_scaling_factor

    def start_camera(self):
        """Start"""
        return

    def set_live(self, mode):
        """Set the live mode.
        Args:
            mode (bool): Live mode.
        self.set_video_live(mode)
        """

    def change_owner(self):
        """LIMA specific, because it has to be root at startup
        move this to Qt4_LimaVideo
        """
        if os.getuid() == 0:
            try:
                os.setgid(int(os.getenv("SUDO_GID")))
                os.setuid(int(os.getenv("SUDO_UID")))
            except Exception:
                logging.getLogger().warning(
                    "%s: failed to change the process ownership.", self.id
                )

    def get_width(self):
        """Get the image width.
        Returns:
           (int): Image width [pixels].
        """
        return int(self.image_dimensions[0])

    def get_height(self):
        """Get the image height.
        Returns:
           (int): Image height [pixels].
        """
        return int(self.image_dimensions[1])

    def do_image_polling(self, sleep_time):
        """
        Descript. :
        """
        while self.get_video_live() is True:
            if USEQT:
                self.get_new_image()
            else:
                self.get_jpg_image()
            time.sleep(sleep_time)

    def connect_notify(self, signal):
        """
        Descript. :
        """
        return

    def refresh_video(self):
        """
        Descript. :
        """

    def set_cam_encoding(self, cam_encoding):
        """Set the image encoding.
        Args:
           cam_encoding(str): set the encoding type.
        """
        if cam_encoding == "yuv422p":
            self.decoder = self.yuv_2_rgb
        elif cam_encoding == "y8":
            self.decoder = self.y8_2_rgb
        elif cam_encoding == "y16":
            self.decoder = self.y16_2_rgb
        elif cam_encoding.lower() == "bayer_rg16":
            self.decoder = self.bayer_rg16_2_rgb
        self.cam_encoding = cam_encoding

    def get_image_dimensions(self):
        """Get the scaled width and the height of the image:
        Returns:
            (list): Width [mm], height [mm] list.
        """
        raw_width, raw_height = self.get_raw_image_size()
        width = raw_width * self.scale
        height = raw_height * self.scale
        return [width, height]

    # -------- Methods to be implemented by the implementing class --------

    def get_raw_image_size(self):
        """Must return a two-value list necessary to avoid breaking
        e.g. ViideoMockup
        """
        return [None, None]

    @abc.abstractmethod
    def get_image(self):
        """The implementing class should return here the latest_image in
        raw_format, followed by the width and height of the image"""

    @abc.abstractmethod
    def get_gain(self):
        """Get the camera gain"""

    @abc.abstractmethod
    def set_gain(self, gain_value):
        """Set the camera gain"""

    @abc.abstractmethod
    def get_exposure_time(self):
        """Get the camera exposure time [s]"""

    @abc.abstractmethod
    def set_exposure_time(self, exposure_time_value):
        """Set the camera exposure time [s]"""

    @abc.abstractmethod
    def get_video_live(self):
        """Get the video live mode.
        Returns:
            flag(bool): Live mode.
        """

    @abc.abstractmethod
    def set_video_live(self, flag):
        """Set the video live mode.
        Args:
            flag(bool): Live mode.
        """

    # Other (no implementation for now. Can be overloaded, otherwise dummy)
    def get_gamma(self):
        """Get the gamma"""
        return

    def set_gamma(self, gamma_value):
        """Set the gamma"""

    def get_contrast(self):
        """Get the contrast"""
        return

    def set_contrast(self, contrast_value):
        """Set the contrast"""

    def get_brightness(self):
        """Get the brightness"""
        return

    def set_brightness(self, brightness_value):
        """Set the brightness"""


def test_hwo(hwo):
    print("Image dimensions: ", hwo.get_image_dimensions())
    print("Live Mode: ", hwo.get_video_live())



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
MicrodiffZoom

Example xml file:
<object class="MicrodiffZoom">
  <username>zoom</username>
  <exporter_address>wid30bmd2s:9001</exporter_address>
  <value_channel_name>CoaxialCameraZoomValue</value_channel_name>
  <state_channel_name>ZoomState</state_channel_name>
  <!-- if levels do not corresponf to values -->
  <values>{"LEVEL1": 1, "LEVEL2": 2, "LEVEL3": 3, "LEVEL4": 4, "LEVEL5": 5, "LEV
EL6": 6}</values>
</object>
"""

from enum import Enum

from mxcubecore.HardwareObjects.ExporterNState import ExporterNState

__copyright__ = """ Copyright © 2020 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class MicrodiffZoom(ExporterNState):
    """MicrodiffZoom class"""

    def init(self):
        """Initialize the zoom"""
        super().init()

        # check if we have values other that UNKNOWN
        _len = len(self.VALUES) - 1
        if _len > 0:
            # we can only assume that the values are consecutive integers
            # so the limits correspond to the keys.
            self.set_limits((1, _len))
        else:
            # no values in the config file, initialise from the hardware.
            self.set_limits(self._get_range())
            self._initialise_values()

    def set_limits(self, limits=(None, None)):
        """Set the low and high limits.
        Args:
            limits (tuple): two integers tuple (low limit, high limit).
        """
        self._nominal_limits = limits

    def update_value(self, value=None):
        """Check if the value has changed. Emits signal valueChanged.
        Args:
            value: value
        """
        # Make sure that update value of super class always is passed value=None
        # so that _get_value is called to get the Enum value and not the numeric
        # value passed by underlying event data.
        super().update_value()

    def update_limits(self, limits=None):
        """Check if the limits have changed. Emits signal limitsChanged.
        Args:
            limits (tuple): two integers tuple (low limit, high limit).
        """
        if not limits:
            limits = self.get_limits()

        # All values are not None nor NaN
        self._nominal_limits = limits
        self.emit("limitsChanged", (limits,))

    def _initialise_values(self):
        """Initialise the ValueEnum from the limits"""
        low, high = self.get_limits()

        values = {f"LEVEL{v}": v for v in range(low, high + 1)}
        self.VALUES = Enum(
            "ValueEnum",
            dict(values, **{item.name: item.value for item in self.VALUES}),
        )

    def _get_range(self):
        """Get the zoom range.
        Returns:
            (tuple): two integers tuple - min and max value.
        """
        try:
            _low, _high = self._exporter.execute("getZoomRange")
        except (AttributeError, ValueError):
            _low, _high = 1, 10

        # inf is a problematic value
        if _low == float("-inf"):
            _low = 1

        if _high == float("inf"):
            _high = 10

        return _low, _high



# pylint: skip-file

import logging
import time

from PyTango import DeviceProxy

from mxcubecore import HardwareRepository as HWR
from mxcubecore.TaskUtils import task


class LimaDetectorMockup:
    def init(self, config, collect_obj=None):
        self.config = config
        self.header = {}

        lima_device = config.get_property("lima_device")
        pilatus_device = config.get_property("pilatus_device")
        if None in (lima_device, pilatus_device):
            return

        for channel_name in (
            "acq_status",
            "acq_trigger_mode",
            "saving_mode",
            "acq_nb_frames",
            "acq_expo_time",
            "saving_directory",
            "saving_prefix",
            "saving_suffix",
            "saving_next_number",
            "saving_index_format",
            "saving_format",
            "saving_overwrite_policy",
            "saving_header_delimiter",
            "last_image_saved",
        ):
            self.add_channel(
                {"type": "tango", "name": channel_name, "tangoname": lima_device},
                channel_name,
            )

        for channel_name in ("fill_mode", "threshold"):
            self.add_channel(
                {"type": "tango", "name": channel_name, "tangoname": pilatus_device},
                channel_name,
            )

        pilatus_tg_device = DeviceProxy(pilatus_device)
        if hasattr(pilatus_tg_device, "working_energy"):
            self.add_channel(
                {
                    "type": "tango",
                    "name": "energy_threshold",
                    "tangoname": pilatus_device,
                },
                "working_energy",
            )
        else:
            self.add_channel(
                {
                    "type": "tango",
                    "name": "energy_threshold",
                    "tangoname": pilatus_device,
                },
                "energy_threshold",
            )

        self.add_command(
            {"type": "tango", "name": "prepare_acq", "tangoname": lima_device},
            "prepareAcq",
        )
        self.add_command(
            {"type": "tango", "name": "start_acq", "tangoname": lima_device}, "startAcq"
        )
        self.add_command(
            {"type": "tango", "name": "stop_acq", "tangoname": lima_device}, "stopAcq"
        )
        self.add_command(
            {"type": "tango", "name": "reset", "tangoname": lima_device}, "reset"
        )
        self.add_command(
            {"type": "tango", "name": "set_image_header", "tangoname": lima_device},
            "SetImageHeader",
        )

    def last_image_saved(self):
        return 2

    def get_deadtime(self):
        return 0.01

    @task
    def prepare_acquisition(
        self,
        take_dark,
        start,
        osc_range,
        exptime,
        npass,
        number_of_images,
        comment,
        energy,
        still,
    ):
        diffractometer_positions = HWR.beamline.diffractometer.get_positions()
        self.start_angles = []
        for i in range(number_of_images):
            self.start_angles.append("%0.4f deg." % (start + osc_range * i))
        kappa_phi = diffractometer_positions.get("kappa_phi", -9999)
        if kappa_phi is None:
            kappa_phi = -9999
        kappa = diffractometer_positions.get("kappa", -9999)
        if kappa is None:
            kappa = -9999

    def set_energy_threshold(self, energy):
        return

    @task
    def set_detector_filenames(
        self, frame_number, start, filename, jpeg_full_path, jpeg_thumbnail_full_path
    ):
        return

    @task
    def start_acquisition(self):
        logging.getLogger("HWR").info("Mockup detector starts acquisition")
        return

    def stop(self):
        time.sleep(1)



"""
Superclass for EPICS actuators.

Should be put as the first superclass,
e.g. class EPICSMotor(EPICSActuator, AbstractMotor):

Example of xml file:

<object class="LNLS.EPICSActuator">
    <channel type="epics" name="epicsActuator_val">MNC:B:LUCIOLE01:LIGHT_CH1</channel>
    <channel type="epics" name="epicsActuator_rbv" polling="500">MNC:B:LUCIOLE01:LIGHT_CH1</channel>
    <username>BackLight</username>
    <motor_name>BackLight</motor_name>
    <default_limits>(0, 8000)</default_limits>
</object>
"""

import random
import time

import gevent

from mxcubecore.HardwareObjects.abstract import AbstractActuator


class EPICSActuator(AbstractActuator.AbstractActuator):
    """EPCIS actuator class"""

    ACTUATOR_VAL = "epicsActuator_val"  # target
    ACTUATOR_RBV = "epicsActuator_rbv"  # readback

    def __init__(self, name):
        super(EPICSActuator, self).__init__(name)
        self.__wait_actuator_task = None
        self._nominal_limits = (-1e4, 1e4)

    def init(self):
        """Initialization method"""
        super(EPICSActuator, self).init()
        self.update_state(self.STATES.READY)

    def _wait_actuator(self):
        """Wait actuator to be ready."""
        time.sleep(0.3)
        self.update_state(self.STATES.READY)

    def get_value(self):
        """Override AbstractActuator method."""
        return self.get_channel_value(self.ACTUATOR_RBV)

    def set_value(self, value, timeout=0):
        """Override AbstractActuator method."""
        if self.read_only:
            raise ValueError("Attempt to set value for read-only Actuator")
        if self.validate_value(value):
            self.update_state(self.STATES.BUSY)
            if timeout or timeout is None:
                with gevent.Timeout(
                    timeout, RuntimeError("Motor %s timed out" % self.username)
                ):
                    self._set_value(value)
                    new_value = self._wait_actuator(value)
            else:
                self._set_value(value)
                self.__wait_actuator_task = gevent.spawn(self._wait_actuator)
        else:
            raise ValueError(
                "Invalid value %s; limits are %s" % (value, self.get_limits())
            )

    def abort(self):
        """Immediately halt movement. By default self.stop = self.abort"""
        if self.__wait_actuator_task is not None:
            self.__wait_actuator_task.kill()
        self.update_state(self.STATES.READY)

    def _set_value(self, value):
        """Override AbstractActuator method."""
        self.set_channel_value(self.ACTUATOR_VAL, value)



from grob import grob_control

from mxcubecore.BaseHardwareObjects import HardwareObject


class Grob(HardwareObject):
    def __init__(self, name):
        super().__init__(name)
        self.SampleTableMotor = grob_control.SampleTableMotor
        self.GonioMotor = grob_control.GonioMotor
        self.SampleMotor = grob_control.SampleMotor

    def init(self):
        self.controller = grob_control.init(self.grob_host, wago_host=self.wago_host)



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
[Name]
ALBAZoomMotor

[Description]
Hardware Object is used to manipulate the zoom of the OAV camera.

[Channels]
- position
- state
- labels

[Commands]

[Emitted signals]
- stateChanged
- predefinedPositionChanged

[Functions]
- None

[Included Hardware Objects]
- None

Example Hardware Object XML file :
==================================
<object class="ALBAZoomMotor">
  <username>Zoom</username>
  <taurusname>ioregister/eh_zoom_tangoior_ctrl/2</taurusname>
  <alias>zoom</alias>
  <actuator_name>Zoom</actuator_name>
  <channel type="sardana" polling="200" name="position">Value</channel>
  <channel type="sardana" polling="200" name="state">State</channel>
  <channel type="sardana" name="labels">Labels</channel>
  <interval>200</interval>
  <threshold>0.001</threshold>
</object>
"""

import logging

import PyTango

from mxcubecore import BaseHardwareObjects
from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor

__author__ = "Bixente Rey"
__credits__ = ["MXCuBE collaboration"]

__version__ = "2.2."
__maintainer__ = "Jordi Andreu"
__email__ = "jandreu[at]cells.es"
__status__ = "Draft"


class ALBAZoomMotor(BaseHardwareObjects.Device, AbstractMotor):
    INIT, FAULT, READY, MOVING, ONLIMIT = range(5)

    def __init__(self, name):
        super().__init__(name)

    def init(self):
        logging.getLogger("HWR").debug("Initializing zoom motor IOR")
        self.positionChannel = self.get_channel_object("position")
        self.stateChannel = self.get_channel_object("state")
        self.labelsChannel = self.get_channel_object("labels")
        self.currentposition = 0
        self.currentstate = None

        self.positionChannel.connect_signal("update", self.positionChanged)
        self.stateChannel.connect_signal("update", self.stateChanged)

    def get_predefined_positions_list(self):
        labels = self.labelsChannel.get_value()
        labels = labels.split()
        retlist = []
        for label in labels:
            #    label, pos = label.split(":")
            #    retlist.append(int(pos))
            pos = str(label.replace(":", " "))
            retlist.append(pos)
        logging.getLogger("HWR").debug("Zoom positions list: %s" % repr(retlist))
        new_retlist = []
        for n, e in enumerate(retlist):
            name = e.split()
            new_retlist.append("%s %s" % (n + 1, name[0]))
        logging.getLogger("HWR").debug("Zoom positions list: %s" % repr(new_retlist))

        # retlist = ["z1 1","z2 2"]
        # logging.getLogger("HWR").debug("Zoom positions list: %s" % repr(retlist))
        return new_retlist

    def moveToPosition(self, posno):
        no = posno.split()[0]
        logging.getLogger("HWR").debug("type %s" % type(no))
        #        no = posno
        logging.getLogger("HWR").debug("Moving to position %s" % no)
        state = self.positionChannel.set_value(int(no))

    def motorIsMoving(self):
        if str(self.get_state()) == "MOVING":
            return True
        else:
            return False

    def get_limits(self):
        return (1, 12)

    def get_state(self):
        state = self.stateChannel.get_value()
        curr_pos = self.get_value()
        if state == PyTango.DevState.ON:
            return ALBAZoomMotor.READY
        elif state == PyTango.DevState.MOVING or state == PyTango.DevState.RUNNING:
            return ALBAZoomMotor.MOVING
        elif curr_pos in self.get_limits():
            return ALBAZoomMotor.ONLIMIT
        else:
            return ALBAZoomMotor.FAULT
        return state

    def get_value(self):
        try:
            return self.positionChannel.get_value()
        except Exception:
            return self.currentposition

    def get_current_position_name(self):
        try:
            n = int(self.positionChannel.get_value())
            value = "%s z%s" % (n, n)
            logging.getLogger("HWR").debug(
                "get_current_position_name: %s" % repr(value)
            )
            return value
        except Exception:
            logging.getLogger("HWR").debug("cannot get name zoom value")
            return None

    def stateChanged(self, state):
        logging.getLogger("HWR").debug("stateChanged emitted: %s" % state)
        the_state = self.get_state()
        if the_state != self.currentstate:
            self.currentstate = the_state
            self.emit("stateChanged", (the_state,))

    def positionChanged(self, currentposition):
        previous_position = self.currentposition
        self.currentposition = self.get_current_position_name()
        if self.currentposition != previous_position:
            logging.getLogger("HWR").debug(
                "predefinedPositionChanged emitted: %s" % self.currentposition
            )
            self.emit("predefinedPositionChanged", (self.currentposition, 0))

    def is_ready(self):
        state = self.get_state()
        return state == ALBAZoomMotor.READY


def test_hwo(zoom):
    print(type(zoom.get_state()))

    print("     Zoom position is : ", zoom.get_value())
    print("Zoom position name is : ", zoom.get_current_position_name())
    print("               Moving : ", zoom.motorIsMoving())
    print("                State : ", zoom.get_state())
    print("            Positions : ", zoom.get_predefined_positions_list())


if __name__ == "__main__":
    test()



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Abstract hardware object for data collection."""

from __future__ import annotations

import abc
import collections
import errno
import logging
import os
import time

import gevent
import gevent.event

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.TaskUtils import task

__credits__ = ["MXCuBE collaboration"]


BeamlineConfig = collections.namedtuple(
    "BeamlineConfig",
    [
        "synchrotron_name",
        "directory_prefix",
        "default_exposure_time",
        "minimum_exposure_time",
        "detector_fileext",
        "detector_type",
        "detector_manufacturer",
        "detector_model",
        "detector_px",
        "detector_py",
        "detector_binning_mode",
        "undulators",
        "focusing_optic",
        "monochromator_type",
        "beam_divergence_vertical",
        "beam_divergence_horizontal",
        "polarisation",
        "input_files_server",
    ],
)


class AbstractCollect(HardwareObject, object):
    """Abstract hardware object for data collection.

    Define a sequence in which data collection is executed.

    Emits:
        collectReady (bool):
        collectOscillationStarted (dict):
        progressInit (tuple[str, int, bool]):
            For example: ``("Collection", progress, bool)``.
        collectOscillationFailed (dict):
        collectOscillationFinished (dict):
        progressStop:
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, name) -> None:
        HardwareObject.__init__(self, name)

        self.bl_config = BeamlineConfig(*[None] * 18)

        self._collecting = False
        self._error_msg = ""
        self.exp_type_dict = {}

        self.collection_id = None
        self.data_collect_task = None
        self.current_dc_parameters = None
        self.current_lims_sample = {}
        self.run_offline_processing = None
        self.run_online_processing = None
        self.ready_event = None

    def init(self):
        self.ready_event = gevent.event.Event()

        undulators = []
        try:
            for undulator in self["undulators"]:
                undulators.append(undulator)
        except BaseException:
            pass

        beam_div_hor, beam_div_ver = HWR.beamline.beam.get_beam_divergence()

        self.set_beamline_configuration(
            synchrotron_name=HWR.beamline.session.synchrotron_name,
            directory_prefix=self.get_property("directory_prefix"),
            default_exposure_time=HWR.beamline.detector.get_property(
                "default_exposure_time"
            ),
            minimum_exposure_time=HWR.beamline.detector.get_property(
                "minimum_exposure_time"
            ),
            detector_fileext=HWR.beamline.detector.get_property("fileSuffix"),
            detector_type=HWR.beamline.detector.get_property("type"),
            detector_manufacturer=HWR.beamline.detector.get_property("manufacturer"),
            detector_model=HWR.beamline.detector.get_property("model"),
            detector_px=HWR.beamline.detector.get_property("px"),
            detector_py=HWR.beamline.detector.get_property("py"),
            detector_binning_mode=HWR.beamline.detector.get_binning_mode(),
            undulators=undulators,
            focusing_optic=self.get_property("focusing_optic"),
            monochromator_type=self.get_property("monochromator"),
            beam_divergence_vertical=beam_div_hor,
            beam_divergence_horizontal=beam_div_ver,
            polarisation=self.get_property("polarisation"),
            input_files_server=self.get_property("input_files_server"),
        )

    def set_beamline_configuration(self, **configuration_parameters) -> None:
        """Sets beamline configuration

        Args:
            configuration_parameters (dict): Configuration parameters.
        """
        self.bl_config = BeamlineConfig(**configuration_parameters)

    def collect(self, owner, dc_parameters_list):
        """Main collection method.

        It spawns "do_collect" as well as ready event handling.

        Args:
            owner (str): Owner instance.
            dc_parameters_list (list[dict]):
                List of dictionary containing all collection parameters.
        """
        self.ready_event.clear()
        self.current_dc_parameters = dc_parameters_list[0]
        self.data_collect_task = gevent.spawn(self.do_collect, owner)
        self.ready_event.wait()
        self.ready_event.clear()
        return self.data_collect_task

    def do_collect(self, owner):
        """Actually collect the sequence.

        Args:
            owner (str): owner instance
        """
        log = logging.getLogger("user_level_log")
        log.info("Collection: Preparing to collect")
        self.emit("collectReady", (False,))
        self.emit(
            "collectOscillationStarted",
            (owner, None, None, None, self.current_dc_parameters, None),
        )
        self.emit("progressInit", ("Collection", 100, False))
        self.collection_id = None

        try:
            # ----------------------------------------------------------------
            # Prepare data collection

            self.open_detector_cover()
            self.open_safety_shutter()
            self.open_fast_shutter()

            # ----------------------------------------------------------------
            # Store information in LIMS

            self.current_dc_parameters["status"] = "Running"
            self.current_dc_parameters["collection_start_time"] = time.strftime(
                "%Y-%m-%d %H:%M:%S"
            )

            logging.getLogger("HWR").info(
                "Collection parameters: %s" % str(self.current_dc_parameters)
            )

            log.info("Collection: Storing data collection in LIMS")
            self.store_data_collection_in_lims()

            log.info(
                "Collection: Creating directories for raw images and processing files"
            )
            self.create_file_directories()

            log.info("Collection: Getting sample info from parameters")
            self.get_sample_info()

            log.info("Collection: Storing sample info in LIMS")
            self.store_sample_info_in_lims()

            if all(
                item is None for item in self.current_dc_parameters["motors"].values()
            ):
                # No centring point defined
                # create point based on the current position
                current_diffractometer_position = (
                    HWR.beamline.diffractometer.get_positions()
                )
                for motor in self.current_dc_parameters["motors"].keys():
                    self.current_dc_parameters["motors"][motor] = (
                        current_diffractometer_position.get(motor)
                    )

            # ----------------------------------------------------------------
            # Move to the centered position and take crystal snapshots

            log.info("Collection: Moving to centred position")
            self.move_to_centered_position()
            self.take_crystal_snapshots()
            self.move_to_centered_position()

            # ----------------------------------------------------------------
            # Set data collection parameters

            if "transmission" in self.current_dc_parameters:
                log.info(
                    "Collection: Setting transmission to %.2f",
                    self.current_dc_parameters["transmission"],
                )
                self.set_transmission(self.current_dc_parameters["transmission"])

            wavelength = self.current_dc_parameters.get("wavelength")
            energy = self.current_dc_parameters.get("energy")
            detector_distance = self.current_dc_parameters.get("detector_distance")
            try:
                resolution = self.current_dc_parameters.get("resolution").get("upper")
            except AttributeError:
                resolution = None

            if wavelength:
                # Wavelength (not having a default) overrides energy
                log.info(
                    "Collection: Setting wavelength to %.4f",
                    self.current_dc_parameters["wavelength"],
                )
                energy = HWR.beamline.energy.calculate_energy(wavelength)
            elif energy:
                log.info(
                    "Collection: Setting energy to %.4f",
                    self.current_dc_parameters["energy"],
                )
                wavelength = HWR.beamline.energy.calculate_wavelength(energy)
            if energy:
                self.set_energy(energy)

            if detector_distance:
                # detector_distance (not having a default) overrides resolution
                resolution = HWR.beamline.resolution.distance_to_resolution(
                    detector_distance, wavelength
                )
                log.info(
                    "Collection: Setting detector distance to %.2f",
                    detector_distance,
                )
                self.set_resolution(resolution)
            elif resolution:
                log.info(
                    "Collection: Setting resolution to %.2f",
                    resolution,
                )
                self.set_resolution(resolution)

            # ----------------------------------------------------------------
            # Site specific implementation of a data collection

            # In order to call the hook with original parameters
            # before update_data_collection_in_lims changes them
            # TODO check why this happens

            self.data_collection_hook()

            # ----------------------------------------------------------------
            # Store information in LIMS

            log.info("Collection: Updating data collection in LIMS")
            self.update_data_collection_in_lims()

        except RuntimeError as e:
            failed_msg = "Data collection failed!\n%s" % str(e)
            self.collection_failed(failed_msg)
        else:
            self.collection_finished()
        finally:
            self.data_collection_cleanup()

    def data_collection_cleanup(self):
        """
        Method called when at end of data collection, successful or not.
        """
        self.close_fast_shutter()
        self.close_safety_shutter()
        self.close_detector_cover()

    def collection_failed(self, failed_msg=None):
        """Collection failed method"""

        if not failed_msg:
            failed_msg = "Data collection failed!"
        self.current_dc_parameters["status"] = "Failed"
        self.current_dc_parameters["comments"] = "%s\n%s" % (
            failed_msg,
            self._error_msg,
        )
        self.emit(
            "collectOscillationFailed",
            (
                None,
                False,
                failed_msg,
                self.current_dc_parameters.get("collection_id"),
                None,
            ),
        )
        self.emit("progressStop", ())
        self._collecting = False
        self.update_data_collection_in_lims()
        self.ready_event.set()

    def collection_stopped(self):
        """Collection stopped method"""

        self.current_dc_parameters["status"] = "Stopped"
        self.current_dc_parameters["comments"] = "Stopped by the user"
        self.emit("progressStop", ())
        self._collecting = False
        self.update_data_collection_in_lims()
        self.ready_event.set()
        if self.data_collect_task is not None:
            self.data_collect_task.kill(block=False)

    def collection_finished(self):
        """Collection finished behaviour"""

        success_msg = "Data collection successful"
        self.current_dc_parameters["status"] = success_msg

        if self.current_dc_parameters["experiment_type"] != "Collect - Multiwedge":
            self.update_data_collection_in_lims()

            last_frame = self.current_dc_parameters["oscillation_sequence"][0][
                "number_of_images"
            ]
            if (
                last_frame > 1
                and self.current_dc_parameters["experiment_type"] != "Mesh"
            ):
                # We do not store first and last frame for mesh
                self.store_image_in_lims_by_frame_num(last_frame)
            if (
                self.current_dc_parameters["experiment_type"] in ("OSC", "Helical")
                and self.current_dc_parameters["oscillation_sequence"][0]["overlap"]
                == 0
                and last_frame > 19
            ):
                self.trigger_auto_processing("after", 0)

        self.emit(
            "collectOscillationFinished",
            (
                None,
                True,
                success_msg,
                self.current_dc_parameters.get("collection_id"),
                None,
                self.current_dc_parameters,
            ),
        )
        self.emit("progressStop", ())
        self._collecting = False
        self.ready_event.set()

    def store_image_in_lims_by_frame_num(self, frame_number):
        pass

    def stop_collect(self):
        """Stop data collection."""
        if self.data_collect_task is not None:
            self.data_collect_task.kill(block=False)

    def open_detector_cover(self):
        """Open detector cover."""

    def open_safety_shutter(self):
        """Open safety shutter."""

    def open_fast_shutter(self):
        """Open fast shutter."""

    def close_fast_shutter(self):
        """Close fast shutter."""

    def close_safety_shutter(self):
        """Close safety shutter."""

    def close_detector_cover(self):
        """Close detector cover."""

    def set_transmission(self, value: float):
        """Set transmission.

        Args:
            value: Transmission value to set.
        """

    def set_wavelength(self, value: float):
        """Set wavelength.

        Args:
            value: Wavelength value to set.
        """

    def set_energy(self, value: float):
        """Set energy.

        Args:
            value: Energy value to set.
        """

    def set_resolution(self, value: float):
        """Set resolution.

        Args:
            value: Resolution value to set.
        """
        HWR.beamline.energy.wait_ready()
        HWR.beamline.resolution.set_value(value)

    def get_total_absorbed_dose(self):
        return

    def get_wavelength(self) -> float:
        """Get current wavelength.

        Returns:
            Current beamline wavelength [Ä].
        """
        if HWR.beamline.energy is not None:
            return HWR.beamline.energy.get_wavelength()

    def get_detector_distance(self) -> float:
        """Get current detector distance.

        Returns:
            Current detector distance position [mm].
        """
        if HWR.beamline.detector is not None:
            return HWR.beamline.detector.distance.get_value()

    def get_resolution(self) -> float:
        """Get current resolution.

        Returns:
            Current resolution [Å].
        """
        if HWR.beamline.resolution is not None:
            return HWR.beamline.resolution.get_value()

    def get_transmission(self) -> float:
        """Get current beamline transmission.

        Returns:
            Current transmission.
        """
        if HWR.beamline.transmission is not None:
            return HWR.beamline.transmission.get_value()

    def get_beam_size(self) -> tuple:
        """Get beam size.

        Returns:
            Beam width, and beam height.
        """
        if HWR.beamline.beam is not None:
            return HWR.beamline.beam.get_beam_size()
        else:
            return None, None

    def get_slit_gaps(self) -> tuple:
        """Get slit gap distance.

        Returns:
            Horizontal gap, and vertical gap.
        """
        if HWR.beamline.beam is not None:
            return HWR.beamline.beam.get_slits_gap()
        return None, None

    def get_undulators_gaps(self) -> dict:
        """Get undulator gaps.

        Returns:
            Undulator gaps.
        """
        return {}
        # return HWR.beamline.energy.get_undulator_gaps()

    def get_machine_current(self) -> float:
        """Get machine current.

        Returns:
            Machine current [mA].
        """
        if HWR.beamline.machine_info:
            return HWR.beamline.machine_info.get_current()
        else:
            return 0

    def get_machine_message(self) -> str:
        """Get machine message.

        Returns:
            Current machine message if any.
        """
        if HWR.beamline.machine_info:
            return HWR.beamline.machine_info.get_message()
        else:
            return ""

    def get_machine_fill_mode(self) -> str:
        """Get machine fill mode.

        Returns:
            Machine filling mode.
        """
        if HWR.beamline.machine_info:
            fill_mode = str(HWR.beamline.machine_info.get_message())
            return fill_mode[:20]
        else:
            return ""

    def get_measured_intensity(self) -> float:
        """Get beam intensity.

        Returns:
            Beam intensity.
        """
        return

    def get_cryo_temperature(self) -> float:
        """Get cryogenic temperature.

        Returns:
            Current cryo temperature [K].
        """
        return

    def create_file_directories(self):
        """Method to create directories for raw and processing files.

        Directory names for xds, mosflm and hkl are created.
        """
        self.create_directories(
            self.current_dc_parameters["fileinfo"]["directory"],
            self.current_dc_parameters["fileinfo"]["process_directory"],
            self.current_dc_parameters["fileinfo"]["archive_directory"],
        )
        xds_directory, mosflm_directory, hkl2000_directory = self.prepare_input_files()
        if xds_directory:
            self.current_dc_parameters["xds_dir"] = xds_directory

    def create_directories(self, *args):
        """Creates folders on disk.

        Args:
            args (list(str)): List of directories to create.
        """
        for directory in args:
            try:
                os.makedirs(directory)
            except os.error as e:
                if e.errno != errno.EEXIST:
                    raise

    def prepare_input_files(self) -> tuple[str | None, str | None, str | None]:
        """Prepare input files for xds, mosflm and hkl2000."""

        return None, None, None

    def store_data_collection_in_lims(self):
        """Store current data collection information in LIMS database."""
        lims = HWR.beamline.lims
        if (
            lims
            and lims.is_connected()
            and not self.current_dc_parameters["in_interleave"]
        ):
            try:
                self.current_dc_parameters["synchrotronMode"] = (
                    self.get_machine_fill_mode()
                )
                (
                    collection_id,
                    detector_id,
                ) = HWR.beamline.lims.store_data_collection(
                    self.current_dc_parameters, self.bl_config
                )
                self.current_dc_parameters["collection_id"] = collection_id
                self.collection_id = collection_id
                if detector_id:
                    self.current_dc_parameters["detector_id"] = detector_id
            except BaseException:
                logging.getLogger("HWR").exception(
                    "Could not store data collection in LIMS"
                )

    def update_data_collection_in_lims(self) -> None:
        """Update current data collection information in LIMS database."""
        params = self.current_dc_parameters
        if HWR.beamline.lims and not params["in_interleave"]:
            params["flux"] = HWR.beamline.flux.get_value()
            params["flux_end"] = params["flux"]
            params["totalAbsorbedDose"] = self.get_total_absorbed_dose()
            params["wavelength"] = HWR.beamline.energy.get_wavelength()
            params["detectorDistance"] = HWR.beamline.detector.distance.get_value()
            params["resolution"] = HWR.beamline.resolution.get_value()
            params["transmission"] = HWR.beamline.transmission.get_value()
            beam_centre_x, beam_centre_y = HWR.beamline.detector.get_beam_position()
            pixel_x, pixel_y = HWR.beamline.detector.get_pixel_size()
            params["xBeam"] = beam_centre_x * pixel_x
            params["yBeam"] = beam_centre_y * pixel_y
            und = self.get_undulators_gaps()
            i = 1
            for jj in self.bl_config.undulators:
                key = jj.type
                if key in und:
                    params["undulatorGap%d" % (i)] = und[key]
                    i += 1
            params["resolutionAtCorner"] = HWR.beamline.resolution.get_value_at_corner()
            beam_size_x, beam_size_y = HWR.beamline.beam.get_beam_size()
            params["beamSizeAtSampleX"] = beam_size_x
            params["beamSizeAtSampleY"] = beam_size_y
            params["beamShape"] = HWR.beamline.beam.get_beam_shape()
            hor_gap, vert_gap = self.get_slit_gaps()
            params["slitGapHorizontal"] = hor_gap
            params["slitGapVertical"] = vert_gap
            try:
                HWR.beamline.lims.update_data_collection(params)
            except BaseException:
                logging.getLogger("HWR").exception(
                    "Could not update data collection in LIMS"
                )

    def store_sample_info_in_lims(self):
        """Store current sample information in LIMS database."""
        lims = HWR.beamline.lims
        if (
            lims
            and lims.is_connected()
            and not self.current_dc_parameters["in_interleave"]
        ):
            HWR.beamline.lims.update_bl_sample(self.current_lims_sample)

    def store_image_in_lims(
        self, frame_number: int, motor_position_id: int | None = None
    ):
        """Store image information in LIMS database.

        Args:
            frame_number: Frame number of the image within the data collection.
            motor_position_id: Motor position ID.
        """
        lims = HWR.beamline.lims
        if (
            lims
            and lims.is_connected()
            and not self.current_dc_parameters["in_interleave"]
        ):
            file_location = self.current_dc_parameters["fileinfo"]["directory"]
            image_file_template = self.current_dc_parameters["fileinfo"]["template"]
            filename = image_file_template % frame_number
            lims_image = {
                "dataCollectionId": self.current_dc_parameters.get("collection_id"),
                "fileName": filename,
                "fileLocation": file_location,
                "imageNumber": frame_number,
                "measuredIntensity": self.get_measured_intensity(),
                "synchrotronCurrent": self.get_machine_current(),
                "machineMessage": self.get_machine_message(),
                "temperature": self.get_cryo_temperature(),
            }
            archive_directory = self.current_dc_parameters["fileinfo"][
                "archive_directory"
            ]
            if archive_directory:
                jpeg_filename = "%s.jpeg" % os.path.splitext(image_file_template)[0]
                thumb_filename = (
                    "%s.thumb.jpeg" % os.path.splitext(image_file_template)[0]
                )
                jpeg_file_template = os.path.join(
                    archive_directory, jpeg_filename
                ).replace("cbf.jpeg", "jpeg")
                jpeg_thumbnail_file_template = os.path.join(
                    archive_directory, thumb_filename
                ).replace("cbf.thumb", "thumb")
                jpeg_full_path = jpeg_file_template % frame_number
                jpeg_thumbnail_full_path = jpeg_thumbnail_file_template % frame_number
                lims_image["jpegFileFullPath"] = jpeg_full_path
                lims_image["jpegThumbnailFileFullPath"] = jpeg_thumbnail_full_path
            if motor_position_id:
                lims_image["motorPositionId"] = motor_position_id
            image_id = HWR.beamline.lims.store_image(lims_image)
            return image_id

    def update_lims_with_workflow(
        self, workflow_id: int, grid_snapshot_filename: str
    ) -> None:
        """Update collection with information about workflow.

        Args:
            workflow_id: Workflow ID.
            grid_snapshot_filename: Grid snapshot file path.
        """
        lims = HWR.beamline.lims
        if lims and lims.is_connected():
            try:
                self.current_dc_parameters["workflow_id"] = workflow_id
                if grid_snapshot_filename:
                    self.current_dc_parameters["xtalSnapshotFullPath3"] = (
                        grid_snapshot_filename
                    )
                HWR.beamline.lims.update_data_collection(self.current_dc_parameters)
            except BaseException:
                logging.getLogger("HWR").exception(
                    "Could not store data collection into ISPyB"
                )

    def get_sample_info(self) -> None:
        """Get current sample information in LIMS database.

        Information is stored in the internal "current_dc_parameters" dictionary.
        """
        sample_info = self.current_dc_parameters.get("sample_reference")
        try:
            sample_id = int(sample_info["blSampleId"])
        except BaseException:
            sample_id = None

        self.current_dc_parameters["blSampleId"] = sample_id

        if HWR.beamline.diffractometer.in_plate_mode():
            # TODO store plate location in lims
            pass
        elif HWR.beamline.sample_changer:
            try:
                self.current_dc_parameters["actualSampleBarcode"] = (
                    HWR.beamline.sample_changer.getLoadedSample().getID()
                )
                self.current_dc_parameters["actualContainerBarcode"] = (
                    HWR.beamline.sample_changer.getLoadedSample().getContainer().getID()
                )

                logging.getLogger("user_level_log").info("Getting loaded sample coords")
                basket, vial = HWR.beamline.sample_changer.getLoadedSample().getCoords()

                self.current_dc_parameters["actualSampleSlotInContainer"] = vial
                self.current_dc_parameters["actualContainerSlotInSC"] = basket
            except BaseException:
                self.current_dc_parameters["actualSampleBarcode"] = None
                self.current_dc_parameters["actualContainerBarcode"] = None
        else:
            self.current_dc_parameters["actualSampleBarcode"] = None
            self.current_dc_parameters["actualContainerBarcode"] = None

    def move_to_centered_position(self) -> None:
        """Move motors to the centered position.

        Move motors to the centered position
        as stored in "current_dc_parameters" dictionary.
        """
        positions_str = ""
        for motor, position in self.current_dc_parameters["motors"].items():
            if position:
                if isinstance(motor, str):
                    positions_str += " %s=%f" % (motor, position)
                else:
                    positions_str += " %s=%f" % (motor.getMotorMnemonic(), position)
        self.current_dc_parameters["actualCenteringPosition"] = positions_str
        self.move_motors(self.current_dc_parameters["motors"])

    @abc.abstractmethod
    @task
    def move_motors(self, motor_position_dict) -> None:
        return

    def take_crystal_snapshots(self) -> None:
        """Take crystal snapshots of the currently loaded sample."""
        number_of_snapshots = self.current_dc_parameters["take_snapshots"]
        if self.current_dc_parameters["experiment_type"] == "Mesh":
            number_of_snapshots = 1

        snapshot_directory = self.current_dc_parameters["fileinfo"]["archive_directory"]
        if number_of_snapshots > 0 or self.current_dc_parameters.get("take_video"):
            if not os.path.exists(snapshot_directory):
                try:
                    self.create_directories(snapshot_directory)
                except BaseException:
                    logging.getLogger("HWR").exception(
                        "Collection: Error creating snapshot directory"
                    )
        if number_of_snapshots > 0 and not self.current_dc_parameters["in_interleave"]:
            logging.getLogger("user_level_log").info(
                "Collection: Taking %d sample snapshot(s)" % number_of_snapshots
            )
            for snapshot_index in range(number_of_snapshots):
                snapshot_filename = os.path.join(
                    snapshot_directory,
                    "%s_%s_%s.snapshot.jpeg"
                    % (
                        self.current_dc_parameters["fileinfo"]["prefix"],
                        self.current_dc_parameters["fileinfo"]["run_number"],
                        (snapshot_index + 1),
                    ),
                )
                self.current_dc_parameters[
                    "xtalSnapshotFullPath%i" % (snapshot_index + 1)
                ] = snapshot_filename
                self._take_crystal_snapshot(snapshot_filename)
                if number_of_snapshots > 1:
                    HWR.beamline.diffractometer.move_omega_relative(90)

        if (
            not HWR.beamline.diffractometer.in_plate_mode()
            and self.current_dc_parameters.get("take_video")
        ):
            # Add checkbox to allow enable/disable creation of gif
            logging.getLogger("user_level_log").info("Collection: Saving animated gif")
            animation_filename = os.path.join(
                snapshot_directory,
                "%s_%s_animation.gif"
                % (
                    self.current_dc_parameters["fileinfo"]["prefix"],
                    self.current_dc_parameters["fileinfo"]["run_number"],
                ),
            )
            self.current_dc_parameters["xtalSnapshotFullPath2"] = animation_filename
            self._take_crystal_animation(animation_filename, duration_sec=1)

    @abc.abstractmethod
    @task
    def _take_crystal_snapshot(self, snapshot_filename):
        """
        Depends on gui version how this method is implemented.
        In Qt3 diffractometer has a function,
        In Qt4 graphics_manager is making crystal snapshots
        """
        pass

    def _take_crystal_animation(self, animation_filename, duration_sec=1):
        """Rotates sample by 360 and composes a gif file"""
        pass

    @abc.abstractmethod
    def data_collection_hook(self):
        """
        Descript. :
        """
        pass

    @abc.abstractmethod
    def trigger_auto_processing(self, process_event, frame_number):
        """
        Descript. :
        """
        pass

    def set_helical(self, arg):
        """
        Descript. :
        """
        pass

    def set_helical_pos(self, arg):
        """
        Descript. :
        """
        pass

    def set_fast_characterisation(self, arg):
        """
        Descript. :
        """
        pass

    def setCentringStatus(self, status):
        """
        Descript. :
        """
        pass

    def prepare_interleave(self, data_model, param_list):
        self.current_dc_parameters = param_list[0]
        self.current_dc_parameters["status"] = "Running"
        self.current_dc_parameters["collection_start_time"] = time.strftime(
            "%Y-%m-%d %H:%M:%S"
        )
        self.take_crystal_snapshots()

        self.store_data_collection_in_lims()
        self.current_dc_parameters["status"] = "Data collection successful"
        self.update_data_collection_in_lims()

    # specifies the next scan will be a mesh scan
    def set_mesh(self, mesh_on):
        self.mesh = mesh_on

    def set_mesh_scan_parameters(
        self,
        num_lines: int,
        total_nb_frames: int,
        mesh_center_param: tuple,
        mesh_range_param: tuple,
    ) -> None:
        """Set the mesh scan parameters."""
        return

        # self.mesh_num_lines = num_lines
        # self.mesh_total_nb_frames = total_nb_frames
        # self.mesh_range = mesh_range_param
        # self.mesh_center = mesh_center_param



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""EMBLSafetyShutter"""

import logging
from enum import (
    Enum,
    unique,
)

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract.AbstractShutter import AbstractShutter

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


@unique
class ShutterValueEnum(Enum):
    """Defines only the compulsory values."""

    OPEN = "Open"
    CLOSED = "Closed"
    UNKNOWN = "Unknown"
    NOPERM = "No permission"
    DISABLED = "Disabled"


class EMBLSafetyShutter(AbstractShutter):
    """
    EMBLSafetyShutter defines interface to DESY ics
    """

    VALUES = ShutterValueEnum

    def __init__(self, name):
        super(EMBLSafetyShutter, self).__init__(name)

        self._nominal_limits = None

        self.use_shutter = None
        self.data_collection_state = None
        self.shutter_can_open = None
        self.shutter_is_open = None
        self.shutter_is_closed = None
        # GB 20190304: per mysteriously disappearing first update of
        # shutter_state_closed:
        self.shutter_can_open = None
        self.ics_enabled = None

        self.chan_collection_state = None
        self.chan_state_open = None
        self.chan_state_closed = None
        self.chan_state_open_permission = None
        self.chan_ics_error = None
        self.chan_error = None
        self.chan_cmd_close_error = None
        self.chan_cmd_open_error = None
        self.cmd_open = None
        self.cmd_close = None

    def init(self):
        super(EMBLSafetyShutter, self).init()

        self.chan_collection_state = self.get_channel_object("chanCollectStatus")
        if self.chan_collection_state:
            self.chan_collection_state.connect_signal(
                "update", self.data_collection_state_changed
            )

        self.chan_state_closed = self.get_channel_object("chanStateClosed")
        self.chan_state_closed.connect_signal("update", self.state_closed_changed)
        self.chan_state_open = self.get_channel_object("chanStateOpen")
        self.chan_state_open.connect_signal("update", self.state_open_changed)

        self.chan_state_open_permission = self.get_channel_object(
            "chanStateOpenPermission"
        )
        self.chan_state_open_permission.connect_signal(
            "update", self.state_open_permission_changed
        )
        self.state_open_permission_changed(self.chan_state_open_permission.get_value())

        self.chan_ics_error = self.get_channel_object("chanIcsError")
        self.chan_ics_error.connect_signal("update", self.ics_error_msg_changed)
        self.ics_error_msg_changed(self.chan_ics_error.get_value())

        self.chan_cmd_close_error = self.get_channel_object("chanCmdCloseError")
        if self.chan_cmd_close_error is not None:
            self.chan_cmd_close_error.connect_signal(
                "update", self.cmd_error_msg_changed
            )

        self.chan_cmd_open_error = self.get_channel_object("chanCmdOpenError")
        if self.chan_cmd_open_error is not None:
            self.chan_cmd_open_error.connect_signal(
                "update", self.cmd_error_msg_changed
            )

        self.cmd_open = self.get_command_object("cmdOpen")
        self.cmd_close = self.get_command_object("cmdClose")

        self.use_shutter = self.get_property("useShutter", True)

        self.state_open_changed(self.chan_state_open.get_value())

    def data_collection_state_changed(self, state):
        """Updates shutter state when data collection state changes

        :param state: data collection state
        :type state: str
        :return: None
        """
        self.data_collection_state = state
        self.update_shutter_state()

    def state_open_changed(self, state):
        """Updates shutter state when shutter open value changes

        :param state: shutter open state
        :type state: str
        :return: None
        """

        self.shutter_is_open = state
        self.update_shutter_state()

    def state_closed_changed(self, state):
        """Updates shutter state when shutter close value changes

        :param state: shutter close state
        :type state: str
        :return: None
        """
        self.shutter_is_closed = state
        self.update_shutter_state()

    def state_open_permission_changed(self, state):
        """Updates shutter state when open permission changes

        :param state: permission state
        :type state: str
        :return: None
        """
        self.shutter_can_open = state
        self.update_shutter_state()

    def cmd_error_msg_changed(self, error_msg):
        """Method called when opening of the shutter fails

        :param error_msg: error message
        :type error_msg: str
        :return: None
        """
        if len(error_msg) > 0:
            logging.getLogger("GUI").error("Safety shutter: Error %s" % error_msg)

    def ics_error_msg_changed(self, error_msg):
        """Updates ICS error message

        :param error_msg: error message
        :type error_msg: str
        :return: None
        """
        if len(error_msg) > 0:
            logging.getLogger("GUI").error("DESY ICS Connection: Error %s" % error_msg)
            self.ics_enabled = False
        else:
            self.ics_enabled = True
        self.update_shutter_state()

    def update_shutter_state(self):
        """Updates shutter state

        :return: shutter state as str
        """
        msg = ""

        if self.data_collection_state == "collecting":
            value = self.VALUES.DISABLED
        elif self.shutter_is_open:
            value = self.VALUES.OPEN
        elif self.shutter_is_closed:
            value = self.VALUES.CLOSED
        elif not self.shutter_can_open:
            value = self.VALUES.DISABLED
        else:
            value = self.VALUES.UNKNOWN

        if not self.shutter_is_open and not self.shutter_can_open:
            value = self.VALUES.NOPERM
            msg = "No permission"

        if not self.ics_enabled:
            value = self.VALUES.DISABLED
            msg = "Ics broke"

        if not self.use_shutter:
            value = self.VALUES.DISABLED

        self.update_value(value)

        return self._nominal_value

    def get_value(self):
        return self._nominal_value

    def control_shutter(self, open_state):
        """Opens or closes shutter

        :param open_state: open state
        :type open_state: bool
        :return: None
        """
        if open_state:
            if self._nominal_value == self.VALUES.CLOSED:
                self.open()
        else:
            if self._nominal_value == self.VALUES.OPEN:
                self.close()

    def _set_value(self, value):
        if value == self.VALUES.OPEN:
            self.cmd_open()
        elif value == self.VALUES.CLOSED:
            self.cmd_close()



#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
XRF Spectrum queue implementation of pre_execute, execute and post_execute
"""

import logging

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.queue_entry.base_queue_entry import (
    QUEUE_ENTRY_STATUS,
    BaseQueueEntry,
    QueueAbortedException,
    QueueExecutionException,
)

__credits__ = ["MXCuBE collaboration"]
__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class XrfSpectrumQueueEntry(BaseQueueEntry):
    """XRF queue handler"""

    def __init__(self, view=None, data_model=None):
        super().__init__(view, data_model)
        self._failed = False

    def __getstate__(self):
        d = dict(self.__dict__)
        d["xrf_spectrum_task"] = None
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)

    def execute(self):
        """Execute"""
        super().execute()
        if HWR.beamline.xrf_spectrum is not None:
            xrf_spectrum = self.get_data_model()
            if xrf_spectrum.shape is not None:
                point = HWR.beamline.sample_view.get_shape(xrf_spectrum.shape)
                xrf_spectrum.centred_position = point.get_centred_position()
            self.get_view().setText(1, "Starting xrf spectrum")
            path_template = xrf_spectrum.path_template
            HWR.beamline.xrf_spectrum.start_spectrum(
                integration_time=xrf_spectrum.count_time,
                data_dir=xrf_spectrum.path_template.directory,
                archive_dir=xrf_spectrum.path_template.get_archive_directory(),
                prefix=f"{path_template.get_prefix()}_{path_template.run_number}",
                session_id=HWR.beamline.session.session_id,
                blsample_id=xrf_spectrum._node_id,
                cpos=xrf_spectrum.centred_position,
            )
            HWR.beamline.xrf_spectrum._ready_event.wait()
            HWR.beamline.xrf_spectrum._ready_event.clear()
        else:
            logging.getLogger("user_level_log").info(
                "XRFSpectrum not defined in beamline setup"
            )
            self.xrf_state_handler(HardwareObjectState.FAULT)

    def pre_execute(self):
        """Pre-execution actions"""
        super().pre_execute()
        self._failed = False
        qctrl = self.get_queue_controller()
        qctrl.connect(
            HWR.beamline.xrf_spectrum,
            "xrfSpectrumStatusChanged",
            self.xrf_spectrum_status_changed,
        )

        qctrl.connect(HWR.beamline.xrf_spectrum, "stateChanged", self.xrf_state_handler)

    def post_execute(self):
        """Post-execution actions"""
        qctrl = self.get_queue_controller()
        qctrl.disconnect(
            HWR.beamline.xrf_spectrum,
            "xrfSpectrumStatusChanged",
            self.xrf_spectrum_status_changed,
        )

        qctrl.disconnect(
            HWR.beamline.xrf_spectrum, "stateChanged", self.xrf_state_handler
        )
        if self._failed:
            raise QueueAbortedException("Queue stopped", self)
        self.get_view().set_checkable(False)
        super().post_execute()

    def xrf_spectrum_status_changed(self, msg):
        """xrfSpectrumStatusChanged handler.
        Args:
            msg (str): Message when xrfSpectrumStatusChanged emitted.
        """
        logging.getLogger("user_level_log").info(msg)

    def xrf_state_handler(self, state=None):
        """State handler - signal connected is stateChanged.
        Args:
            state (enum): HardwareObjectState enum member.
        Raises:
             QueueExecutionException: If procedure failed.
        """
        state = state or HWR.beamline.xrf_spectrum.get_state()
        if state == HWR.beamline.xrf_spectrum.STATES.BUSY:
            logging.getLogger("user_level_log").info("XRF spectrum started.")
            self.get_view().setText(1, "In progress")
        if state == HWR.beamline.xrf_spectrum.STATES.READY:
            logging.getLogger("user_level_log").info("XRF spectrum finished.")
            self.get_view().setText(1, "Done")
        if state == HWR.beamline.xrf_spectrum.STATES.FAULT:
            self._failed = True
            self.get_view().setText(1, "Failed")
            self.status = QUEUE_ENTRY_STATUS.FAILED
            logging.getLogger("user_level_log").error("XRF spectrum failed.")
            raise QueueExecutionException("XRF spectrum failed", self)

    def get_type_str(self):
        return "XRF spectrum"



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import subprocess
import time

import gevent
import numpy as np
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure

from mxcubecore import HardwareRepository as HWR
from mxcubecore import TaskUtils
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractEnergyScan import AbstractEnergyScan

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


class EMBLEnergyScan(AbstractEnergyScan, HardwareObject):
    def __init__(self, name):
        AbstractEnergyScan.__init__(self)
        HardwareObject.__init__(self, name)
        self._tunable_bl = True

        self.status = None
        self.startup_done = False
        self.ready_event = None
        self.scanning = False
        self.th_edge = None
        self.scan_directory = None
        self.scan_data = None
        self.scan_prefix = None
        self.num_points = None
        self.scan_info = None
        self.chooch_cmd = None

        self.chan_scan_start = None
        self.chan_scan_status = None
        self.chan_scan_error = None

        self.cmd_scan_abort = None
        self.cmd_adjust_transmission = False
        self.cmd_set_max_transmission = False

    def init(self):
        self.ready_event = gevent.event.Event()
        self.scan_info = {}

        self.chan_scan_start = self.get_channel_object("energyScanStart")
        self.chan_scan_start.connect_signal("update", self.scan_start_update)
        self.chan_scan_status = self.get_channel_object("energyScanStatus")
        self.chan_scan_status.connect_signal("update", self.scan_status_update)
        self.chan_scan_error = self.get_channel_object("energyScanError")
        self.chan_scan_error.connect_signal("update", self.scan_error_update)

        self.cmd_scan_abort = self.get_command_object("energyScanAbort")
        self.cmd_adjust_transmission = self.get_command_object("cmdAdjustTransmission")
        self.cmd_set_max_transmission = self.get_command_object("cmdSetMaxTransmission")

        self.num_points = self.get_property("numPoints", 60)
        self.chooch_cmd = self.get_property("chooch_command")

    def scan_start_update(self, values):
        """Emits new scan point

        :param values: list of values
        :type values: list of two floats
        :return:
        """
        if self.scanning:
            self.emit_new_data_point(values)

    def scan_status_update(self, status):
        if self.scanning and status != self.status:
            if status == "scanning":
                logging.getLogger("GUI").info("Energy scan: Executing...")

                if HWR.beamline.transmission is not None:
                    self.scan_info["transmissionFactor"] = (
                        HWR.beamline.transmission.get_value()
                    )
                else:
                    self.scan_info["transmissionFactor"] = None
            elif status == "ready":
                if self.scanning is True:
                    logging.getLogger("GUI").info("Energy scan: Finished")
                    self.scanCommandFinished()
            elif status == "aborting":
                if self.scanning is True:
                    self.scanCommandAborted()
                    logging.getLogger("GUI").info("Energy scan: Aborted")
            elif status == "error":
                self.scanCommandFailed()
            self.status = status

    def scan_error_update(self, error_msg):
        """Prints error message

        :param error_msg: error message
        :type error_msg: str
        :return: None
        """
        if len(error_msg) > 0 and self.startup_done:
            logging.getLogger("GUI").error("Energy scan: %s" % error_msg)

    def emit_new_data_point(self, values):
        """Adds new point to the energy scan curve

        :param values: values
        :type values: list of two floats
        :return: None
        """
        if len(values) > 0:
            if type(values) in (tuple, list):
                if type(values[-1]) not in (tuple, list):
                    values = [values]
                x = values[-1][0]
                y = values[-1][1]
                # if x is in keV, transform into eV otherwise let it like it is
                # if point larger than previous point (for chooch)
                if x > 0 and y > 0:
                    if len(self.scan_data) > 0:
                        if x > self.scan_data[-1][0]:
                            self.scan_data.append([(x < 1000 and x * 1000.0 or x), y])
                    else:
                        self.scan_data.append([(x < 1000 and x * 1000.0 or x), y])
                    # a = str(x < 1000 and x*1000.0 or x) + " -- " + str(y)
                    # logging.getLogger("GUI").info(a)
                    self.emit("scanNewPoint", ((x < 1000 and x * 1000.0 or x), y))
                    self.emit("progressStep", (len(self.scan_data)))

    def is_connected(self):
        return True

    def start_energy_scan(
        self,
        element,
        edge,
        directory,
        prefix,
        session_id=None,
        blsample_id=None,
        exptime=3,
    ):
        """Starts energy scan

        :param element: scan element
        :type element: str
        :param edge: edge
        :type edge: str
        :param directory: scan directory
        :type directory: str
        :param prefix: scan prefix
        :type prefix: str
        :param session_id: session id
        :type session_id: int
        :param blsample_id: sample id
        :type blsample_id: int
        :param exptime: exposure time in seconds
        :type exptime: float
        :return: True if success, otherwise returns False
        """
        log = logging.getLogger("HWR")

        self.scan_info = {
            "sessionId": session_id,
            "blSampleId": blsample_id,
            "element": element,
            "edgeEnergy": edge,
        }
        self.scan_data = []
        self.scan_directory = directory
        self.scan_prefix = prefix
        self.startup_done = True

        """
        if not os.path.isdir(directory):
            log.debug("EnergyScan: creating directory %s" % directory)
            try:
                os.makedirs(directory)
            except OSError as diag:
                log.error(
                    "EnergyScan: error creating directory %s (%s)"
                    % (directory, str(diag))
                )
                self.emit("energyScanStatusChanged", ("Error creating directory",))
                return False
        """

        if self.chan_scan_status.get_value() in ["ready", "unknown", "error"]:
            if hasattr(HWR.beamline.energy, "release_break_bragg"):
                HWR.beamline.energy.release_break_bragg()

            # if self.transmission_hwobj is not None:
            #    self.scan_info['transmissionFactor'] = self.transmission_hwobj.get_value()
            # else:
            #    self.scan_info['transmissionFactor'] = None
            self.scan_info["exposureTime"] = exptime
            self.scan_info["startEnergy"] = 0
            self.scan_info["endEnergy"] = 0
            self.scan_info["startTime"] = str(time.strftime("%Y-%m-%d %H:%M:%S"))
            size_hor = None
            size_ver = None
            if HWR.beamline.beam is not None:
                size_hor, size_ver = HWR.beamline.beam.get_beam_size()
                size_hor = size_hor * 1000
                size_ver = size_ver * 1000
            self.scan_info["beamSizeHorizontal"] = size_hor
            self.scan_info["beamSizeVertical"] = size_ver
            self.chan_scan_start.set_value("%s;%s" % (element, edge))
            self.scanCommandStarted()
        else:
            log.error(
                "Another energy scan in progress. "
                + "Please wait when the scan is finished"
            )
            self.emit(
                "energyScanStatusChanged",
                (
                    "Another energy scan in progress. "
                    + "Please wait when the scan is finished"
                ),
            )
            self.scanCommandFailed()

            return False

        return True

    def cancelEnergyScan(self, *args):
        if self.scanning:
            self.cmd_scan_abort()
            self.scanCommandAborted()

    def scanCommandStarted(self, *args):
        if self.scan_info["blSampleId"]:
            title = "Sample: %s Element: %s Edge: %s" % (
                self.scan_info["blSampleId"],
                self.scan_info["element"],
                self.scan_info["edgeEnergy"],
            )
        else:
            title = "Element: %s Edge: %s" % (
                self.scan_info["element"],
                self.scan_info["edgeEnergy"],
            )

        graph_info = {
            "xlabel": "energy",
            "ylabel": "counts",
            "scaletype": "normal",
            "title": title,
        }
        self.scanning = True
        self.emit("energyScanStarted", graph_info)
        self.emit("progressInit", "Energy scan", self.num_points, False)

    def scanCommandFailed(self, *args):
        with TaskUtils.cleanup(self.ready_event.set):
            # error_msg = self.chan_scan_error.get_value()
            # print error_msg
            # logging.getLogger("GUI").error("Energy scan: %s" % error_msg)
            self.scan_info["endTime"] = str(time.strftime("%Y-%m-%d %H:%M:%S"))
            if self.scan_data:
                self.scan_info["startEnergy"] = self.scan_data[-1][0] / 1000.0
                self.scan_info["endEnergy"] = self.scan_data[-1][1] / 1000.0
            self.emit("energyScanFailed", ())
            self.emit("progressStop", ())

            if hasattr(HWR.beamline.energy, "set_break_bragg"):
                HWR.beamline.energy.set_break_bragg()
            self.scanning = False
            self.ready_event.set()

    def scanCommandAborted(self, *args):
        self.emit("energyScanFailed", ())
        self.emit("progressStop", ())
        if hasattr(HWR.beamline.energy, "set_break_bragg"):
            HWR.beamline.energy.set_break_bragg()
        self.scanning = False
        self.ready_event.set()

    def scanCommandFinished(self, *args):
        with TaskUtils.cleanup(self.ready_event.set):
            self.scan_info["endTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
            logging.getLogger("HWR").debug("Energy scan: finished")
            self.scanning = False
            self.scan_info["startEnergy"] = self.scan_data[-1][0]
            self.scan_info["endEnergy"] = self.scan_data[-1][1]
            self.emit("energyScanFinished", (self.scan_info,))
            self.emit("progressStop", ())
            if hasattr(HWR.beamline.energy, "set_break_bragg"):
                HWR.beamline.energy.set_break_bragg()

    def do_chooch(self, elt, edge, scan_directory, archive_directory, prefix):
        archive_file_prefix = str(os.path.join(archive_directory, prefix))

        if os.path.exists(archive_file_prefix + ".raw"):
            i = 1
            while os.path.exists(archive_file_prefix + "%d.raw" % i):
                i = i + 1
            archive_file_prefix += "_%d" % i

        archive_file_raw_filename = os.path.extsep.join((archive_file_prefix, "raw"))
        archive_file_efs_filename = os.path.extsep.join((archive_file_prefix, "efs"))
        archive_file_png_filename = os.path.extsep.join((archive_file_prefix, "png"))

        try:
            if not os.path.exists(archive_directory):
                os.makedirs(archive_directory)
        except Exception:
            logging.getLogger("HWR").exception(
                "EMBLEnergyScan: could not create results directory."
            )
            self.store_energy_scan()
            self.emit("energyScanFailed", ())
            return

        try:
            archive_file_raw = open(archive_file_raw_filename, "w")
        except Exception:
            logging.getLogger("HWR").exception(
                "EMBLEnergyScan: could not create results raw file"
            )
            self.store_energy_scan()
            self.emit("energyScanFailed", ())
            return
        else:
            scanData = []
            x_array = []
            y_array = []
            for i in range(len(self.scan_data)):
                x = float(self.scan_data[i][0])
                x = x < 1000 and x * 1000.0 or x
                y = float(self.scan_data[i][1])
                scanData.append((x, y))
                x_array.append(x / 1000.0)
                y_array.append(y)
                archive_file_raw.write("%f,%f\r\n" % (x, y))
            archive_file_raw.close()
            self.scan_info["scanFileFullPath"] = str(archive_file_raw_filename)

        try:
            p = subprocess.Popen(
                [self.chooch_cmd, archive_file_raw_filename, elt, edge],
                stdout=subprocess.PIPE,
            )

            chooch_results_list = p.communicate()[0].split("\n")
            chooch_results_list.remove("")
            pk, fppPeak, fpPeak, ip, fppInfl, fpInfl = map(
                float, chooch_results_list[-2].split(" ")
            )
            chooch_graph_data = eval(chooch_results_list[-1])
        except Exception:
            self.store_energy_scan()

            logging.getLogger("GUI").error("Energy scan: Chooch failed")
            return None, None, None, None, None, None, None, [], [], [], None

        rm = (pk + 30) / 1000.0
        pk = pk / 1000.0
        savpk = pk
        ip = ip / 1000.0
        comm = ""
        self.scan_info["edgeEnergy"] = 0.1
        self.th_edge = self.scan_info["edgeEnergy"]
        logging.getLogger("GUI").info(
            "Energy Scan: Chooch results are pk=%.2f, ip=%.2f, rm=%.2f" % (pk, ip, rm)
        )

        self.scan_info["peakEnergy"] = pk
        self.scan_info["inflectionEnergy"] = ip
        self.scan_info["remoteEnergy"] = rm
        self.scan_info["peakFPrime"] = fpPeak
        self.scan_info["peakFDoublePrime"] = fppPeak
        self.scan_info["inflectionFPrime"] = fpInfl
        self.scan_info["inflectionFDoublePrime"] = fppInfl
        self.scan_info["comments"] = comm
        self.scan_info["choochFileFullPath"] = archive_file_efs_filename
        self.scan_info["filename"] = archive_file_raw_filename
        self.scan_info["workingDirectory"] = archive_directory

        chooch_graph_x, chooch_graph_y1, chooch_graph_y2 = zip(*chooch_graph_data)
        chooch_graph_x = list(chooch_graph_x)
        for i in range(len(chooch_graph_x)):
            chooch_graph_x[i] = chooch_graph_x[i] / 1000.0

        # logging.getLogger("HWR").info("EMBLEnergyScan: Saving png" )
        # prepare to save png files
        title = "%s  %s  %s\n%.4f  %.2f  %.2f\n%.4f  %.2f  %.2f" % (
            "energy",
            "f'",
            "f''",
            pk,
            fpPeak,
            fppPeak,
            ip,
            fpInfl,
            fppInfl,
        )
        fig = Figure(figsize=(15, 11))
        ax = fig.add_subplot(211)
        ax.set_title("%s\n%s" % (archive_file_efs_filename, title))
        ax.grid(True)
        ax.plot(x_array, y_array, **{"color": "black"})
        ax.set_xlabel("Energy (keV)")
        ax.set_ylabel("MCA counts")
        ax.set_xticklabels(
            np.round(
                np.linspace(
                    min(x_array), max(x_array), len(ax.get_xticklabels()), endpoint=True
                ),
                3,
            )
        )
        ax2 = fig.add_subplot(212)
        ax2.grid(True)
        ax2.set_xlabel("Energy (keV)")
        ax2.set_ylabel("")
        handles = []
        handles.append(ax2.plot(chooch_graph_x, chooch_graph_y1, color="blue"))
        handles.append(ax2.plot(chooch_graph_x, chooch_graph_y2, color="red"))
        ax2.set_xticklabels(
            np.round(
                np.linspace(
                    min(x_array), max(x_array), len(ax.get_xticklabels()), endpoint=True
                ),
                3,
            )
        )
        ax2.axvline(pk, linestyle="--", color="blue")
        ax2.axvline(ip, linestyle="--", color="red")

        canvas = FigureCanvasAgg(fig)

        self.scan_info["jpegChoochFileFullPath"] = str(archive_file_png_filename)
        try:
            logging.getLogger("HWR").info(
                "Saving energy scan to archive directory for ISPyB : %s",
                archive_file_png_filename,
            )
            canvas.print_figure(archive_file_png_filename, dpi=80)
        except Exception:
            logging.getLogger("HWR").exception("could not save figure")

        self.store_energy_scan()

        self.emit(
            "choochFinished",
            (
                pk,
                fppPeak,
                fpPeak,
                ip,
                fppInfl,
                fpInfl,
                rm,
                chooch_graph_x,
                chooch_graph_y1,
                chooch_graph_y2,
                title,
            ),
        )
        return (
            pk,
            fppPeak,
            fpPeak,
            ip,
            fppInfl,
            fpInfl,
            rm,
            chooch_graph_x,
            chooch_graph_y1,
            chooch_graph_y2,
            title,
        )

    def scan_status_changed(self, status):
        self.emit("energyScanStatusChanged", (status,))

    def updateEnergyScan(self, scan_id, jpeg_scan_filename):
        pass

    def get_elements(self):
        elements = []
        try:
            for el in self["elements"]:
                elements.append(
                    {
                        "symbol": el.get_property("symbol"),
                        "energy": el.get_property("energy"),
                    }
                )
        except IndexError:
            pass
        return elements

    def get_scan_data(self):
        """Returns energy scan data.
        List contains tuples of (energy, counts)
        """
        return self.scan_data

    def store_energy_scan(self):
        if HWR.beamline.lims:
            db_status = HWR.beamline.lims.storeEnergyScan(self.scan_info)

    def adjust_transmission(self, state):
        """
        Enables/disables usage of maximal transmission set
        during the energy scan
        """
        self.cmd_adjust_transmission(state)

    def set_max_transmission(self, value):
        """
        Sets maximal transmission used during the energy scan
        """
        self.cmd_set_max_transmission(value)

    def get_adjust_transmission_state(self):
        return self.cmd_adjust_transmission.get()

    def get_max_transmission_value(self):
        return self.cmd_set_max_transmission.get()
